/***************************************************************************
                          contact.cpp  -  description
                             -------------------
    begin                : Sun Jan 5 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contact.h"

#include <qdir.h>
#include <qpixmap.h>

#include <kdebug.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kstddirs.h>

#include "contactlistviewitem.h"
#include "contactextension.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"
#include "../specialgroups.h"

#ifdef KMESSDEBUG_CONTACT
//  #define KMESSDEBUG_CONTACT_GENERAL
//  #define KMESSDEBUG_CONTACT_OFFLINE
#endif

// The constructor
Contact::Contact(QString handle, QString friendlyName, QString list, QString groupId)
 : ContactBase(handle, friendlyName),
   allowed_(false),
   blocked_(false),
   friend_(false),
   groupId_(groupId),
   initialized_(false),
   item_(0),
   reverse_(false),
   status_("FLN")
{
#ifdef KMESSTEST
  bool goodGroupId;
  groupId.toInt( &goodGroupId );
  ASSERT( goodGroupId || groupId == "" );
  ASSERT( ( list == "FL" ) || ( list == "AL" ) || ( list == "BL" ) || ( list == "RL" ) );
#endif
  if ( list == "FL" )
  {
    setFriend( true );
  }
  else if ( list == "AL" )
  {
    setAllowed( true );
  }
  else if ( list == "BL" )
  {
    setBlocked( true );
  }
  else if ( list == "RL" )
  {
    setReverse( true );
  }

  extension_ = new ContactExtension( handle );

  connect ( extension_, SIGNAL(     changedFriendlyName() ),
            this,         SLOT( emitChangedFriendlyName() ) );
  connect ( extension_, SIGNAL(          changedPicture() ),
            this,         SLOT(   forwardChangedPicture() ) );
}



// The constructor
Contact::Contact(QString handle, QString friendlyName, int list, QString groupId)
 : ContactBase(handle, friendlyName),
   allowed_(false),
   blocked_(false),
   friend_(false),
   groupId_(groupId),
   initialized_(false),
   item_(0),
   reverse_(false),
   status_("FLN")
{
#ifdef KMESSTEST
  bool goodGroupId;
  groupId.toInt( &goodGroupId );
  ASSERT( goodGroupId || groupId == "" );
#endif
  if ( list & 1 )
  {
    friend_ = true;
  }
  if ( list & 2 )
  {
    allowed_ = true;
  }
  if ( list & 4 )
  {
    blocked_ = true;
  }
  if ( list & 8 )
  {
    reverse_ = true;
  }

  extension_ = new ContactExtension( handle );

  connect ( extension_, SIGNAL(     changedFriendlyName() ),
            this,         SLOT( emitChangedFriendlyName() ) );
  connect ( extension_, SIGNAL(          changedPicture() ),
            this,         SLOT(   forwardChangedPicture() ) );
}



// The destructor
Contact::~Contact()
{
  delete extension_;
#ifdef KMESSDEBUG_CONTACT_GENERAL
  kdDebug() << "DESTROYED Contact " << friendlyName_ << " (" << handle_ << ") " << endl;
#endif
}



// Check if this is a new contact
bool Contact::checkIfContactAddedUser()
{
#ifdef KMESSDEBUG_CONTACT_GENERAL
  kdDebug() << "Contact " << friendlyName_ << " (" << handle_ << ") - Check if new." << endl;
#endif
  // Check if the contact is a new contact
  if ( !isFriend() && !isBlocked() && !isAllowed() && isReverse() )
  {
    return true;
  }
  return false;
}



// Slot to pass on the changedFriendlyName()
void Contact::emitChangedFriendlyName()
{
  updateItemIdentifierAndPixmap();

  emit changedFriendlyName();
}



// Forward the "changed picture" signal
void Contact::forwardChangedPicture()
{
  emit changedPicture();
}



// Return the path to the contact's picture
QString Contact::getContactPicturePath() const
{
  QString picturePath;

  picturePath = extension_->getContactPicturePath();

  if ( picturePath.isEmpty() )
  {
   // Not found, get a placeholder
   KStandardDirs  directory;
   // Return a path to the "unknown contact" picture.
   picturePath = directory.findResource( "data", "kmess/pics/unknown.png" );
  }

  return picturePath;
}



// Return a pointer to the extension class
ContactExtension *Contact::getExtension() const
{
  return extension_;
}


// Return the contact's friendly name
QString Contact::getFriendlyName() const
{
  if ( extension_->getUseAlternativeName() )
  {
    return extension_->getAlternativeName();
  }
  else
  {
    return friendlyName_;
  }
}



// Return a pointer to the contact's list view item
ContactListViewItem* Contact::getItem()
{
  return item_;
}



// Return a group ID based on the user's visibility and mode settings.
QString Contact::getMetaGroupId() const
{
  QString         metaGroupId;
  CurrentAccount  *currentAccount;

  currentAccount = CurrentAccount::instance();
  if ( currentAccount == 0 )
  {
    kdDebug() << "Contact - getMetaGroupId() - Couldn't get an instance of the current contact." << endl;
    return groupId_;
  }
  // First, check if the user is viewing contacts by group or by online/offline
  if ( currentAccount->getShowContactsByGroup() )
  {
    // The user is showing contacts by group, so the contact's meta group id is the
    //  same as its real group id in the case of friends, or the removed or allowed
    //  groups depending on the contact's list affiliation
    if ( isFriend() )
    {
      metaGroupId = groupId_;
    }
    else if ( isAllowed() )
    {
      metaGroupId = SpecialGroups::ALLOWED;
    }
    else
    {
      metaGroupId = SpecialGroups::REMOVED;
    }
  }
  else
  {
    // The user is showing contacts by online or offline status.  The group of the contact
    //  then depends on whether or not the contact is online.
    if ( getStatus() != "FLN" )
    {
      metaGroupId = SpecialGroups::ONLINE;
    }
    else
    {
      metaGroupId = SpecialGroups::OFFLINE;
    }
  }
  return metaGroupId;
}



// Return the group that the contact's in
QString Contact::getParentGroupId() const
{
  return parentGroupId_;
}



// Return the contact's status
QString Contact::getStatus() const
{
  return status_;
}



// Return the contact's true friendly name, regardless of extension
QString Contact::getTrueFriendlyName() const
{
  return friendlyName_;
}



// Initialize the object
bool Contact::initialize(ContactListViewItem *groupItem, const QString parentGroupId)
{
  if ( initialized_ )
  {
    kdDebug() << "Contact " << getHandle() << " already initialized." << endl;
    return false;
  }
  setParentGroupId( parentGroupId );
  item_ = new ContactListViewItem( groupItem );
  item_->setDragEnabled( true );
  //item_->setDropEnabled( true );
  //item_->setRenameEnabled( 0, true );
  updateItemIdentifierAndPixmap();
  updateItemVisibility();
  initialized_ = true;
  return true;
}



// Whether or not the contact is allowed
bool Contact::isAllowed() const
{
  return allowed_;
}



// Whether or not the contact is blocked
bool Contact::isBlocked() const
{
  return blocked_;
}



// Whether or not the contact is on the friends list
bool Contact::isFriend() const
{
  return friend_;
}



// Whether or not the contact is on the reverse list
bool Contact::isReverse() const
{
  return reverse_;
}



// Set whether or not the contact is allowed
void Contact::setAllowed(bool allowed)
{
  allowed_ = allowed;
  emit changedList();
#ifdef KMESSTEST
  if ( allowed_ )
  {
    ASSERT( !blocked_ );
  }
#endif
}



// Set whether or not the contact is blocked
void Contact::setBlocked(bool blocked)
{
  blocked_ = blocked;
#ifdef KMESSTEST
  if ( blocked_ )
  {
    ASSERT( !allowed_ );
  }
#endif
}



// Set whether or not the contact is on the friends list
void Contact::setFriend(bool isFriend)
{
  friend_ = isFriend;
  emit changedList();
}



// Change the contact's friendly name
void Contact::setFriendlyName(QString newName)
{
  if ( newName != "" )
  {
    friendlyName_ = newName;
    updateItemIdentifierAndPixmap();
    emit changedFriendlyName();
  }
}



// Change the contact's normal group id
void Contact::setGroupId(QString groupId)
{
  groupId_ = groupId;
}



// Set whether or not the user is on a given list
void Contact::setList(const QString list, bool isMember)
{
#ifdef KMESSTEST
  ASSERT( ( list == "FL" ) || ( list == "AL" ) || ( list == "BL" ) || ( list == "RL" ) );
#endif
  if ( list == "FL" )
  {
    setFriend( isMember );
  }
  else if ( list == "AL" )
  {
    setAllowed( isMember );
  }
  else if ( list == "BL" )
  {
    setBlocked( isMember );
    updateItemIdentifierAndPixmap();
  }
  else if ( list == "RL" )
  {
    setReverse( isMember );
    updateItemIdentifierAndPixmap();
  }
  updateItemVisibility();
  // Check if the contact is a new contact
  if ( !isFriend() && !isBlocked() && !isAllowed() && isReverse() )
  {
    // This contact is *only* on the reverse list and so is new
    emit contactAddedUser( handle_, friendlyName_ );
  }
}



// Set whether or not the user is on a given list
void Contact::setList(int list)
{
  if ( list & 1 )
  {
    friend_ = true;
  }
  else
  {
    friend_ = false;
  }
  if ( list & 2 )
  {
    allowed_ = true;
  }
  else
  {
    allowed_ = false;
  }
  if ( list & 4 )
  {
    blocked_ = true;
  }
  else
  {
    blocked_ = false;
  }
  if ( list & 8 )
  {
    reverse_ = true;
  }
  else
  {
    reverse_ = false;
  }
  updateItemIdentifierAndPixmap();
  updateItemVisibility();
  // Check if the contact is a new contact
  if ( !isFriend() && !isBlocked() && !isAllowed() && isReverse() )
  {
    // This contact is *only* on the reverse list and so is new
    emit contactAddedUser( handle_, friendlyName_ );
  }
}



// Tell the contact what group it's in
void Contact::setParentGroupId(const QString parentGroupId)
{
#ifdef KMESSTEST
  bool goodParentGroupId;
  parentGroupId.toInt( &goodParentGroupId );
  ASSERT( goodParentGroupId );
#endif
  parentGroupId_ = parentGroupId;
}



// Set whether or not the contact is on the reverse list
void Contact::setReverse(bool reverse)
{
  reverse_ = reverse;
  emit changedList();
}



// Set the contact's status
void Contact::setStatus(const QString status)
{
#ifdef KMESSTEST
  ASSERT(    ( status == "AWY" ) || ( status == "BRB" ) || ( status == "BSY" ) || ( status == "FLN" )
          || ( status == "IDL" ) || ( status == "LUN" ) || ( status == "NLN" ) || ( status == "PHN" ) );
#endif
  if (    ( status == "AWY" ) || ( status == "BRB" ) || ( status == "BSY" )
       || ( status == "FLN" ) || ( status == "IDL" ) || ( status == "LUN" )
       || ( status == "NLN" ) || ( status == "PHN" ) )
  {
    QString oldStatus = status_;

    status_ = status;
    updateItemIdentifierAndPixmap();
    updateItemVisibility();
    emit changedStatus();
#ifdef KMESSDEBUG_CONTACT_GENERAL
    kdDebug() << "Contact " << friendlyName_ << " went from " << oldStatus << " to " << status_ << "." << endl;
#endif
    // Check if the status changed from online to offline
    if( ( status_ == "FLN" ) && ( oldStatus != "FLN" ) )
    {
#ifdef KMESSDEBUG_CONTACT_GENERAL
      kdDebug() << "Contact went offline." << endl;
#endif
      emit contactOffline( this );
    }
    // Check if the status changed from offline to online
    else if( ( status_ != "FLN" ) && ( oldStatus == "FLN" ) )
    {
#ifdef KMESSDEBUG_CONTACT_GENERAL
      kdDebug() << "Contact went online." << endl;
#endif
      emit contactOnline( this );
    }
  }
}



// Update the item identifier and pixmap
void Contact::updateItemIdentifierAndPixmap()
{
  KIconLoader  load;
  QString      iconName = "online", identifier, onlineIdentifier, sortIdentifier, statusIdentifier = "";
  unsigned int onlineStatus = 1;

  // Go through each possible status_ and set the icon to load and status_ identifier (like "(Away)" )
  if ( blocked_ )
  { // If the contact is blocked, then regardless of status, the identifier "(Blocked)" goes after the name
    //  and the blocked icon is shown.
    statusIdentifier = i18n( "Blocked" );
    iconName = "blocked";
    onlineStatus = ( status_ != "FLN" ) ? 2 : 0;
  }
  else if ( status_ == "AWY" )
  {
    statusIdentifier = i18n( "Away" );
    iconName = "away";
  }
  else if ( status_ == "BRB" )
  {
    statusIdentifier = i18n( "Be Right Back" );
    iconName = "berightback";
  }
  else if ( status_ == "BSY" )
  {
    statusIdentifier = i18n( "Busy" );
    iconName = "busy";
  }
  else if ( status_ == "FLN" )
  {
    statusIdentifier = "";
    iconName = "offline";
    onlineStatus = 0;
  }
  else if ( status_ == "IDL" )
  {
    statusIdentifier = i18n( "Away - Idle" );
    iconName = "away";
  }
  else if ( status_ == "LUN" )
  {
    statusIdentifier = i18n( "Out to Lunch" );
    iconName = "lunch";
  }
  else if ( status_ == "NLN" )
  {
    statusIdentifier = "";
    iconName = "online";
    onlineStatus = 2;
  }
  else if ( status_ == "PHN" )
  {
    statusIdentifier = i18n( "On the Phone" );
    iconName = "onthephone";
  }
  else
  {
    kdDebug() << "Contact - '" << status_ << "' is not a valid status for contact " << handle_ << "." << endl;
    return;
  }

  item_->setItalic( !isReverse() );
  item_->setIdentifier( getFriendlyName() );
  item_->setStatusIdentifier( statusIdentifier );

  //item_->setText( SpecialColumns::SORT,        sortIdentifier   );
  item_->setOnlineStatus( onlineStatus );

  item_->parent()->sortChildItems( 0, false );

  item_->setIcon( iconName );
}



// Update the visibilitiy of the item
void Contact::updateItemVisibility()
{
  CurrentAccount  *currentAccount;
  bool             visible;

  // Visibility depends on whether or not allowed and removed contacts are shown.
  currentAccount = CurrentAccount::instance();
  if ( currentAccount == 0 )
  {
    kdDebug() << "Contact - updateVisibility() - Couldn't get an instance of the current contact." << endl;
    return;
  }
  // If the contact is offline and offline contacts aren't shown..
  if ( ( getStatus() == "FLN" ) && ( !currentAccount->getShowOfflineContacts() ) )
  {
    visible = false;
  }
  else
  { 
    if ( isFriend() )
    {
      // Friends should always be visible
      visible = true;
    }
    else if ( isAllowed() )
    {
      // Whether or not allowed contacts are visible depends on the contact setting
      visible = currentAccount->getShowAllowedContacts();
    }
    else
    {
      // Whether or not removed contacts are visible depends on the contact setting
      visible = currentAccount->getShowRemovedContacts();
    }
  }
#ifdef KMESSDEBUG_CONTACT_OFFLINE
  kdDebug() << "Contact - Show Allowed = " << currentAccount->getShowAllowedContacts() << " Show Offline = " << currentAccount->getShowOfflineContacts() << " Show Removed = " << currentAccount->getShowRemovedContacts() << "." << endl;
  kdDebug() << "Contact " << getFriendlyName() << " FL = " << isFriend() << " AL = " << isAllowed() << " BL = " << isBlocked() << " RL = " << isReverse() << " Visible = " << visible << "." << endl;
#endif
  item_->setVisible( visible );
}
