/***************************************************************************
                          contactlist.cpp  -  description
                             -------------------
    begin                : Sun Jan 5 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "contactlist.h"

#include <qlistview.h>
#include <qtooltip.h>

#include <kaction.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <krun.h>
#include <kurl.h>

#include "../dialogs/contactaddeduserdialog.h"
#include "../dialogs/contactnamedialog.h"
#include "../dialogs/contactpropertiesdialog.h"
#include "../dialogs/groupnamedialog.h"
#include "contact.h"
#include "contactextension.h"
#include "contactlistview.h"
#include "contactlistviewitem.h"
#include "contactlisttooltip.h"
#include "../currentaccount.h"
#include "group.h"
#include "../kmessdebug.h"
#include "../specialgroups.h"

#ifdef KMESSDEBUG_CONTACTLIST
//  #define KMESSDEBUG_CONTACTLIST_GENERAL
  #define KMESSDEBUG_CONTACTLIST_TOOLTIP
//  #define KMESSDEBUG_CONTACTLIST_DISCONNECTION
#endif

// The constructor
ContactList::ContactList()
 : initialized_(false)
{
#ifdef KMESSTEST
  ASSERT( contacts_.count() == 0 );
  ASSERT( groups_.count() == 0 );
#endif
}



// The destructor
ContactList::~ContactList()
{
  Contact *contact;
  Group   *group;

  // Delete all contacts
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    delete contact;
  }
  contacts_.clear();

  // Delete all groups
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    delete group;
  }
  groups_.clear();
  // Delete "removed" groups
  for ( group = removedGroups_.first(); group; group = removedGroups_.next() )
  {
    delete group;
  }
  removedGroups_.clear();

#ifdef KMESSTEST
  ASSERT( contacts_.count() == 0 );
  ASSERT( groups_.count() == 0 );
#endif
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "DESTROYED ContactList " << endl;
#endif
}



// Add a contact to the contact list
void ContactList::addContact( QString handle, QString friendlyName, QString list, QString groupId )
{
  Contact *contact;
  Group   *group;
  bool     initialized;
  QString metaGroupId;

  if ( !handle.isEmpty() )
  {
    // Create the contact
    contact = new Contact( handle, friendlyName, list, groupId );
    // Have the contact calculate a meta group id
    metaGroupId = contact->getMetaGroupId();
    // Make sure that the contact's parent group exists
    group = getGroupById( metaGroupId );
    if ( group == 0 )
    {
      kdDebug() << "ContactList - addContact - Contact " << handle << "'s parent group, " << groupId << " doesn't exist." << endl;
      return;
    }
    // Initialize the contact
    initialized = contact->initialize( group->getItem(), metaGroupId );
    if ( !initialized )
    {
      kdDebug() << "ContactList - The contact " << handle << " couldn't be initialized." << endl;
      delete contact;
      return;
    }
    // Connect the contact to signal if it's a new contact
    connect( contact, SIGNAL(      contactAddedUser(QString, QString) ),
             this,    SLOT  (      contactAddedUser(QString, QString) ) );
    // Connect the contact to signal if it goes online or offline
    connect( contact, SIGNAL(        contactOffline(const Contact*)     ),
             this,    SLOT  ( forwardContactOffline(const Contact*)     ) );
    connect( contact, SIGNAL(         contactOnline(const Contact*)     ),
             this,    SLOT  (  forwardContactOnline(const Contact*)     ) );

    // Update its group's identifier
    group->updateItemIdentifier();
#ifdef KMESSTEST
    ASSERT( contact != 0 );
    uint oldContactCount = contacts_.count();
#endif
    // Add the contact to the list of contacts
    contacts_.append( contact );
#ifdef KMESSTEST
    ASSERT( contacts_.count() == ( oldContactCount + 1 ) );
#endif
    // Check if the contact added the user
    if ( contact->checkIfContactAddedUser() )
    {
      contactAddedUser( contact->getHandle(), contact->getFriendlyName() );
    }
  }
}



// Add a contact to the contact list
void ContactList::addContact( QString handle, QString friendlyName, int list, QString groupId )
{
  Contact *contact;
  Group   *group;
  bool     initialized;
  QString metaGroupId;

  if ( !handle.isEmpty() )
  {
    // Create the contact
    contact = new Contact( handle, friendlyName, list, groupId );
    // Have the contact calculate a meta group id
    metaGroupId = contact->getMetaGroupId();
    // Make sure that the contact's parent group exists
    group = getGroupById( metaGroupId );
    if ( group == 0 )
    {
      kdDebug() << "ContactList - addContact - Contact " << handle << "'s parent group, " << groupId << " doesn't exist." << endl;
      return;
    }
    // Initialize the contact
    initialized = contact->initialize( group->getItem(), metaGroupId );
    if ( !initialized )
    {
      kdDebug() << "ContactList - The contact " << handle << " couldn't be initialized." << endl;
      delete contact;
      return;
    }
    // Connect the contact to signal if it's a new contact
    connect( contact, SIGNAL(      contactAddedUser(QString, QString) ),
             this,    SLOT  (      contactAddedUser(QString, QString) ) );
    // Connect the contact to signal if it goes online or offline
    connect( contact, SIGNAL(        contactOffline(const Contact*)     ),
             this,    SLOT  ( forwardContactOffline(const Contact*)     ) );
    connect( contact, SIGNAL(         contactOnline(const Contact*)     ),
             this,    SLOT  (  forwardContactOnline(const Contact*)     ) );

    // Update its group's identifier
    group->updateItemIdentifier();
#ifdef KMESSTEST
    ASSERT( contact != 0 );
    uint oldContactCount = contacts_.count();
#endif
    // Add the contact to the list of contacts
    contacts_.append( contact );
#ifdef KMESSTEST
    ASSERT( contacts_.count() == ( oldContactCount + 1 ) );
#endif
    // Check if the contact added the user
    if ( contact->checkIfContactAddedUser() )
    {
      contactAddedUser( contact->getHandle(), contact->getFriendlyName() );
    }
  }
}



// Add a group to the contact list
void ContactList::addGroup( QString groupId, QString groupName )
{
  Group *group;
  // Check if the group already exists
  if ( groupExists( groupId ) )
  {
    kdDebug() << "ContactList - addGroup - Group " << groupId << " already exists." << endl;
    return;
  }
  // Create the group
  group = new Group( groupId, groupName, contactListView_ );
  group->plugAction( moveContactToGroup_ );
  connect( group, SIGNAL(            actionSelected(QString) ),
           this,  SLOT  ( moveCurrentContactToGroup(QString) ) );
#ifdef KMESSTEST
  ASSERT( group != 0 );
  uint oldGroupCount = groups_.count();
#endif
  // Add it to the list of groups
  groups_.append( group );
#ifdef KMESSTEST
  ASSERT( groups_.count() == ( oldGroupCount + 1 ) );
#endif
  // Sort the contact list
  contactListView_->sort();
}



// Change a contact's status (and update the friendly name)
void ContactList::changeContactStatus( QString handle, QString friendlyName, QString status )
{
  Contact *contact;
  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - changeContactStatus - Couldn't find contact " << handle << " in the contact list." << endl;
    return;
  }
  contact->setFriendlyName( friendlyName );
  contact->setStatus( status );
  checkContactGroup( contact );
}



// Check that a contact is in the correct group
void ContactList::checkContactGroup( Contact *contact, bool updateGroups  )
{
  QString   currentGroupId, desiredGroupId;
  Group     *currentGroup,  *desiredGroup;

  currentGroupId = contact->getParentGroupId();
  // Get the group the contact is in
  currentGroup = getGroupById( currentGroupId );
  if ( currentGroup == 0 )
  {
    kdDebug() << "ContactList - checkContactGroup - Couldn't get current group " << currentGroup << "." << endl;
    return;
  }

  desiredGroupId = contact->getMetaGroupId();
  // If the contact is not in the group specified by its meta group id...
  if ( currentGroupId != desiredGroupId )
  {
    desiredGroup = getGroupById( desiredGroupId );
    if ( desiredGroup == 0 )
    {
      kdDebug() << "ContactList - checkContactGroup - Couldn't get desired group " << currentGroup << "." << endl;
      return;
    }
    // Take the contact item from its current group
    currentGroup->getItem()->takeItem( contact->getItem() );
    // Insert it into its desired group.
    desiredGroup->getItem()->insertItem( contact->getItem() );
    contact->setParentGroupId( desiredGroupId );
    if ( updateGroups )
    {
      currentGroup->updateItemIdentifier();
      desiredGroup->updateItemIdentifier();
    }
  }
  else
  {
    // Update the current group, anyway
    if ( updateGroups )
    {
      currentGroup->updateItemIdentifier();
    }
  }
}



// Collapse a group item
void ContactList::collapseGroup( QListViewItem *item )
{
  /*KIconLoader load;
  item->setPixmap( 0, load.loadIcon( "2rightarrow", KIcon::Small ) );*/
  ((ContactListViewItem *) item)->setIcon( "2rightarrow" );
}



// A contact has added this user to their contact list.  Show a dialog on
//  whether or not to add this contact.
void ContactList::contactAddedUser(QString handle, QString friendlyName)
{
  // Note that this function isn't well tested because the case where
  //  somebody adds you to their contact list doesn't occur very often,
  //  not when you've been using messenger for a while.
  ContactAddedUserDialog *dialog;
  int                     returnValue;

  dialog = new ContactAddedUserDialog();

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "Contact added user" << endl;
#endif
  // Show a dialog and let the user choose whether to add, allow,
  //  or block the contact.
  returnValue = dialog->launch(handle, friendlyName);
  delete dialog;
  if ( returnValue == ContactAddedUserDialog::ADD )
  { // Add the contact to the allowed and friends lists.
    emit requestAddContact(handle, "AL", "0");
    emit requestAddContact(handle, "FL", "0");
  }
  else if ( returnValue == ContactAddedUserDialog::ALLOW )
  { // Add the contact to the allowed list.
    emit requestAddContact(handle, "AL", "0");
  }
  else if ( returnValue == ContactAddedUserDialog::BLOCK )
  { // Add the contact to the blocked list.
    emit requestAddContact(handle, "BL", "0");
  }
}



// Return whether or not the contact with the given handle exists
bool ContactList::contactExists( QString handle )
{
  Contact *contact;
  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    return false;
  }
  else
  {
    return true;
  }
}



// Set a contact to offline
void ContactList::contactOffline( QString handle )
{
  Contact *contact;
  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - contactOffline - Couldn't find contact " << handle << " in the contact list." << endl;
    return;
  }
  contact->setStatus( "FLN" );
  checkContactGroup( contact );
}



// Delete all the contacts
void ContactList::deleteAllContacts()
{
  Contact       *contact;
  Group         *group;
  QListViewItem *item, *child, *nextChild;

#ifdef KMESSDEBUG_CONTACTLIST_DISCONNECTION
  kdDebug() << "ContactList - Deleting all contacts." << endl;
#endif

  // Go through all the groups and take all their items from them
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    item = group->getItem();
    child = item->firstChild();
    while ( child )
    {
#ifdef KMESSDEBUG_CONTACTLIST_DISCONNECTION
      kdDebug() << "ContactList - Taking " << child->text(0) << " from " << item->text(0) << "." << endl;
#endif
      // Get the child's sibling while the child's still valid
      nextChild = child->nextSibling();
      // Remove the child item from the group
      item->takeItem( child );
      // Delete the child?

      // Go to the next child
      child = nextChild;
    }
  }
  // Delete all contacts
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    delete contact;
  }
  contacts_.clear();
}



// Delete all non-special groups
void ContactList::deleteNonSpecialGroups()
{
  Group     *group;
  QString   id;
  QPtrList<Group> groupsToDelete;

#ifdef KMESSDEBUG_CONTACTLIST_DISCONNECTION
  kdDebug() << "ContactList - Deleting non-special groups." << endl;
#endif

  // Delete all groups
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    id = group->getId();
    if (    ( id != SpecialGroups::ONLINE )
         && ( id != SpecialGroups::OFFLINE )
         && ( id != SpecialGroups::ALLOWED )
         && ( id != SpecialGroups::REMOVED ) )
    {
      groupsToDelete.append( group );
    }
  }
  // Remove the groups to delete from the groups
  for ( group = groupsToDelete.first(); group; group = groupsToDelete.next() )
  {
    contactListView_->takeItem( group->getItem() );
    group->unplugAction( moveContactToGroup_ );
    groups_.remove( group );
    delete group;
  }
#ifdef KMESSTEST
  ASSERT( groups_.count() == 4 );
#endif
}



// Using hotmail's "compose", send an email to the current contact.
void ContactList::emailCurrentContact()
{
  QString contactHandle;

  // Get the current contact
  contactHandle = getContactHandleByItem( contactListView_->currentItem() );
  if ( contactHandle != "" )
  {
    currentAccount_->openHotmailAtCompose( contactHandle );
  }
}



// Expand a group item
void ContactList::expandGroup( QListViewItem *item )
{
  /*KIconLoader load;
  item->setPixmap( 0, load.loadIcon( "2downarrow", KIcon::Small ) );*/
  ((ContactListViewItem *) item)->setIcon( "2downarrow" );
}



// Forward from a contact that it is offline
void ContactList::forwardContactOffline(const Contact *contact)
{
#ifdef KMESSTEST
  ASSERT( contact != 0 );
#endif
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  if ( contact != 0 )
  {
    kdDebug() << "ContactList - Forwarding that " << contact->getFriendlyName() << " went offline." << endl;
  }
#endif
  emit contactOffline( contact );
}



// Forward from a contact that it is online
void ContactList::forwardContactOnline(const Contact *contact)
{
#ifdef KMESSTEST
  ASSERT( contact != 0 );
#endif
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  if ( contact != 0 )
  {
    kdDebug() << "ContactList - Forwarding that " << contact->getFriendlyName() << " went online." << endl;
  }
#endif
  emit contactOnline( contact );
}



// Return the contact with the given handle
Contact* ContactList::getContactByHandle( QString handle )
{
  Contact *contact;
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    if ( contact->getHandle() == handle )
    {
      return contact;
    }
  }
  return 0;
}



// Return the contact with the given list view item
Contact* ContactList::getContactByItem( QListViewItem *item )
{
  Contact *contact;
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    if ( contact->getItem() == item )
    {
      return contact;
    }
  }
  return 0;
}



// Get the friendly name of the given contact
QString ContactList::getContactFriendlyNameByHandle( const QString& handle )
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "Looking for the friendly name of " << handle << endl;
#endif
  Contact *contact;
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    if ( contact->getHandle() == handle )
    {
      return contact->getFriendlyName();
    }
  }
  return "";
}



// Return the handle of the contact with the given list view item
QString ContactList::getContactHandleByItem( QListViewItem *item )
{
  Contact *contact;
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    if ( contact->getItem() == item )
    {
      return contact->getHandle();
    }
  }
  return "";
}



// Return the list of contacts
const QPtrList<Contact>& ContactList::getContactList() const
{
  return contacts_;
}



// Get a pointer to the contact list dragged
Contact* ContactList::getContactLastDragged()
{
  QListViewItem *item;
  Contact *contact;

  item = (QListViewItem *) contactListView_->getLastItemDragged();
  //if (item == 0) return 0;

  contact = getContactByItem( item );
  
  return contact;
}



// Return the group with the given id
Group* ContactList::getGroupById( QString groupId )
{
#ifdef KMESSTEST
  bool goodNumber;
  groupId.toInt( &goodNumber );
  ASSERT( goodNumber );
#endif
  Group *group;
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    if ( group->getId() == groupId )
    {
      return group;
    }
  }
  return 0;
}



// Return the group with the given list view item
Group* ContactList::getGroupByItem( QListViewItem *item )
{
  Group *group;
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    if ( group->getItem() == item )
    {
      return group;
    }
  }
  return 0;
}



// Return the group with the given sort value
Group* ContactList::getGroupBySortPosition( int sortPosition )
{
  Group *group;
  for ( group = groups_.first(); group; group = groups_.next() )
  {
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
    kdDebug() << "ContactList - Does group " << group->getName() << "'s sort position (" << group->getSortPosition() << ") = " << sortPosition <<  "?" << endl;
#endif
    if ( group->getSortPosition() == sortPosition )
    {
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
      kdDebug() << "ContactList - Yes." << endl;
#endif
      return group;
    }
  }
  return 0;
}



// Return whether or not a group with the given id exists
bool ContactList::groupExists( QString groupId )
{
  Group *group;
  group = getGroupById( groupId );
  if ( group == 0 )
  {
    return false;
  }
  else
  {
    return true;
  }
}



// Initialize the contact list
bool ContactList::initialize(ContactListView *contactListView)
{

#ifdef KMESSTEST
  ASSERT( contacts_.count() == 0 );
  ASSERT( groups_.count() == 0 );
#endif
  if ( initialized_ )
  {
    kdDebug() << "ContactList already initialized." << endl;
    return false;
  }
  // Save the instance of the current account
  currentAccount_ = CurrentAccount::instance();
  if ( currentAccount_ == 0 )
  {
    kdDebug() << "ContactList couldn't get an instance of the current account." << endl;
    return false;
  }
  // Connect the current account to signal changes in "show allowed" and "removed" contacts and in the view mode
  connect( currentAccount_, SIGNAL( changedDisplaySettings() ),
           this,            SLOT  (     updateVisibilities() ) );
  connect( currentAccount_, SIGNAL(        changedViewMode() ),
           this,            SLOT  (         updateViewMode() ) );

  // Store the contact list
  if ( contactListView == 0 )
  {
    kdDebug() << "ContactList initialized with null contact list view." << endl;
    return false;
  }
  contactListView_ = contactListView;
	contactListView_->setSorting( SpecialColumns::SORT, true );
  contactListView_->setDragAutoScroll( true );
  // Connect the contact list up so opening and closing groups causes the group items to expand and collapse.
  connect( contactListView_, SIGNAL(                   expanded(QListViewItem*)                              ),
           this,             SLOT  (                expandGroup(QListViewItem*)                              ) );
  connect( contactListView_, SIGNAL(                  collapsed(QListViewItem*)                              ),
           this,             SLOT  (              collapseGroup(QListViewItem*)                              ) );
  connect( contactListView_, SIGNAL(       contextMenuRequested(QListViewItem*, const QPoint&, int)          ),
           this,             SLOT  (           itemRightClicked(QListViewItem*, const QPoint&, int)          ) );
  connect( contactListView_, SIGNAL(              doubleClicked(QListViewItem*)                              ),
           this,             SLOT  (     requestChatWithContact(QListViewItem*)                              ) );
  connect( contactListView_, SIGNAL(                itemRenamed(QListViewItem*, int, const QString&)         ),
           this,             SLOT  (                itemRenamed(QListViewItem*, int, const QString&)         ) );
  connect( contactListView_, SIGNAL(           moveGroupRequest(ContactListViewItem *, ContactListViewItem *)),
           this,             SLOT  (    moveGivenContactToGroup(ContactListViewItem *, ContactListViewItem *)) );
  // Initialize the popups
  initContactActionPopup();
  initGroupActionPopup();
  // Initialize the tooltip
  toolTip_ = new ContactListToolTip( contactListView_, this );
  // Create the special groups
  addGroup( SpecialGroups::ONLINE,  i18n("Online")  );
  addGroup( SpecialGroups::OFFLINE, i18n("Offline") );
  addGroup( SpecialGroups::ALLOWED, i18n("Allowed") );
  addGroup( SpecialGroups::REMOVED, i18n("Removed") );
#ifdef KMESSTEST
  ASSERT( contacts_.count() == 0 );
  ASSERT( groups_.count() == 4 );
#endif
  return true;
}



// Initialize the menu that pops up when a contact is right-clicked.
void ContactList::initContactActionPopup()
{
  // Initialize the KActions used in the contact popup menu
  addContact_ =        new KAction(i18n("&Add Contact"),               "bookmark_add", 0, 0, "add");
  allowContact_ =      new KAction(i18n("&Allow Contact"),             "bookmark_add", 0, 0, "allow");
  blockContact_ =      new KAction(i18n("&Block Contact"),             "cancel",       0, 0, "block");
  chatWithContact_ =   new KAction(i18n("&Chat"),                      "history",      0, 0, "chat");
  emailContact_ =      new KAction(i18n("&Send email"),                "mail_generic", 0, 0, "email");
  contactProperties_ = new KAction(i18n("&Properties"),                "contents",     0, 0, "properties");
  removeContact_ =     new KAction(i18n("&Remove Contact"),            "eraser",       0, 0, "remove");
  msnProfile_ =        new KAction(i18n("&Show Profile"),              "kmess",        0, 0, "profile");
  unblockContact_ =    new KAction(i18n("&Unblock Contact"),           "cancel",       0, 0, "unblock");

  // Connect the actions
  connect( addContact_,       SIGNAL( activated()                     ),
           this,              SLOT  ( requestAddContact()             ) );
  connect( allowContact_,     SIGNAL( activated()                     ),
           this,              SLOT  ( requestAllowContact()           ) );
  connect( blockContact_,     SIGNAL( activated()                     ),
           this,              SLOT  ( requestBlockContact()           ) );
  connect( chatWithContact_,  SIGNAL( activated()                     ),
           this,              SLOT  ( requestChatWithCurrentContact() ) );
  connect( contactProperties_,SIGNAL( activated()                     ),
           this,              SLOT  ( showContactProperties()         ) );
  connect( emailContact_,     SIGNAL( activated()                     ),
           this,              SLOT  ( emailCurrentContact()           ) );
  connect( msnProfile_,       SIGNAL( activated()                     ),
           this,              SLOT  ( showContactMsnProfile()         ) );
  connect( removeContact_,    SIGNAL( activated()                     ),
           this,              SLOT  ( requestRemoveContact()          ) );
  connect( unblockContact_,   SIGNAL( activated()                     ),
           this,              SLOT  ( requestUnblockContact()         ) );


  // Initialize KActionMenus used in the contact popup menu
  moveContactToGroup_ = new KActionMenu(i18n("&Move to Group"), "move", 0, "moveToGroup");

  // Initialize the popup menu
  contactActionPopup_ = new KPopupMenu(this);
  contactActionPopup_->setTitle("KMess");

  // Plug the actions into their menus
  contactProperties_-> plug(contactActionPopup_);
  contactActionPopup_->insertSeparator();
  blockContact_->      plug(contactActionPopup_);
  unblockContact_->    plug(contactActionPopup_);
  contactActionPopup_->insertSeparator();
  addContact_->        plug(contactActionPopup_);
  allowContact_->      plug(contactActionPopup_);
  removeContact_->     plug(contactActionPopup_);
  contactActionPopup_->insertSeparator();
  emailContact_->      plug(contactActionPopup_);
  msnProfile_->        plug(contactActionPopup_);
  chatWithContact_->   plug(contactActionPopup_);
  contactActionPopup_->insertSeparator();
  moveContactToGroup_->plug(contactActionPopup_);
}



// Initialize the menu that pops up when a contact is right-clicked.
void ContactList::initGroupActionPopup()
{
  // Initialize the KActions used in the contact popup menu
  addGroup_      = new KAction(i18n("&Add Group"),       "bookmark_add", 0, 0, "add"      );
  moveGroupDown_ = new KAction(i18n("Move Group &Down"), "1downarrow",   0, 0, "movedown" );
  moveGroupUp_   = new KAction(i18n("Move Group &Up"),   "1uparrow",     0, 0, "moveup"   );
  removeGroup_   = new KAction(i18n("Re&move Group"),    "cancel",       0, 0, "remove"   );
  renameGroup_   = new KAction(i18n("Re&name Group"),    "bookmark",     0, 0, "rename"   );

  // Connect the actions
  connect( addGroup_,      SIGNAL( activated()            ),
           this,           SLOT  ( newGroupDialog()       ) );
  connect( moveGroupDown_, SIGNAL( activated()            ),
           this,           SLOT  ( moveCurrentGroupDown() ) );
  connect( moveGroupUp_,   SIGNAL( activated()            ),
           this,           SLOT  ( moveCurrentGroupUp()   ) );
  connect( removeGroup_,   SIGNAL( activated()            ),
           this,           SLOT  ( removeCurrentGroup()   ) );
  connect( renameGroup_,   SIGNAL( activated()            ),
           this,           SLOT  ( renameCurrentGroup()   ) );

  // Initialize the popup menu
  groupActionPopup_ = new KPopupMenu(this);
  groupActionPopup_->setTitle("KMess");

  // Plug the actions into their menus
  addGroup_->     plug(groupActionPopup_);
  removeGroup_->  plug(groupActionPopup_);
  renameGroup_->  plug(groupActionPopup_);
  groupActionPopup_->insertSeparator();
  moveGroupUp_->  plug(groupActionPopup_);
  moveGroupDown_->plug(groupActionPopup_);
}



// An item in the contact list was renamed using in-place renaming.  We must decide what to do...
void ContactList::itemRenamed(QListViewItem *item, int /*column*/, const QString& text)
{
  Group   *group;
  Contact *contact;
  ContactExtension *extension;
  bool     doRename;

#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "New name " << text << endl;
#endif

  group   =   getGroupByItem( item );
  contact = getContactByItem( item );

  // If a contact was renamed...
  if ( contact != 0 )
  {
    extension = contact->getExtension();

    extension->setUseAlternativeName( true );
    extension->setAlternativeName( text );
  }
  else if ( group != 0 )
  {
    doRename = ( text != group->getName() );

    if ( doRename )
    {
      emit requestRenameGroup( group->getId(), text );
    }
  }
}



// An item in the contact list was right-clicked, so bring up a menu
void ContactList::itemRightClicked(QListViewItem *item, const QPoint& point, int)
{
  const Group   *group;
  const Contact *contact;

  group   =   getGroupByItem( item );
  contact = getContactByItem( item );

  // If a contact was right clicked...
  if ( contact != 0 )
  {
    // The contact can be blocked only if it's not on the blocked list,
    //  unblocked only if it is.
    if ( contact->isBlocked() )
    {
      blockContact_->setEnabled(false);
      unblockContact_->setEnabled(true);
    }
    else
    {
      blockContact_->setEnabled(true);
      unblockContact_->setEnabled(false);
    }
    // The contact can be chatted with only if it's on the friends list
    //  but not on the blocked list.
    if ( ( contact->isFriend() ) && ( !contact->isBlocked() ) )
    {
      // And only if the contact is online
      if ( contact->getStatus() != "FLN" )
      {
        chatWithContact_->setEnabled(true);
      }
      else
      {
        chatWithContact_->setEnabled(false);
      }
    }
    else
    {
      chatWithContact_->setEnabled(false);
    }
    // Contacts can be removed and change groups as long as they're on the
    //  friends list, added only if they're not.
    if ( contact->isFriend() )
    {
      addContact_->setEnabled(false);
      removeContact_->setEnabled(true);
      moveContactToGroup_->setEnabled(true);
      allowContact_->setEnabled(false);
    }
    else
    {
      addContact_->setEnabled(true);
      removeContact_->setEnabled(false);
      moveContactToGroup_->setEnabled(false);
      if ( ( !contact->isAllowed() ) &&  ( !contact->isBlocked() ) )
      { // An item can be "allowed" only if it is on only the removed list.
        allowContact_->setEnabled(true);
      }
      else
      {
        allowContact_->setEnabled(false);
      }
    }
    // Show the popup
    contactActionPopup_->popup(point);
  }
  else if ( group != 0 )
  {
    // If the group is not a special group...
    if (   ( group->getId() != SpecialGroups::ALLOWED )
        && ( group->getId() != SpecialGroups::REMOVED )
        && ( group->getId() != SpecialGroups::ONLINE  )
        && ( group->getId() != SpecialGroups::OFFLINE ) )
    {
      // Show the popup
      groupActionPopup_->popup(point);
    }
  }
}



// Change a contact in the contact list or add it if it doesn't exist
void ContactList::modifyContact( QString handle, QString friendlyName, QString list, QString groupId )
{
  Contact *contact;
  contact = getContactByHandle( handle );
  // If the contact doesn't exist, create it.
  if ( contact == 0 )
  {
    addContact( handle, friendlyName, list, groupId );
    return;
  }
  // Otherwise, modify the existing contact;
  contact->setList( list, true );
  if ( list == "FL" )
  {
    contact->setGroupId( groupId );
  }
  checkContactGroup( contact );
}



// Change a contact in the contact list or add it if it doesn't exist
void ContactList::modifyContact( QString handle, QString friendlyName, int list, QString groupId )
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - Modify " << handle << endl;
#endif
  Contact *contact;
  contact = getContactByHandle( handle );
  // If the contact doesn't exist, create it.
  if ( contact == 0 )
  {
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - " << handle << " doesn't exist, so create him." << endl;
#endif
    addContact( handle, friendlyName, list, groupId );
    return;
  }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - " << handle << " was found.  Change his list or whatever." << endl;
#endif
  // Otherwise, modify the existing contact;
  contact->setList( list );
  if ( list & 1 )
  {
    contact->setGroupId( groupId );
  }
  checkContactGroup( contact );
}



// Move the current contact to the given group.
void ContactList::moveCurrentContactToGroup(QString groupId)
{
  const Contact *contact;

  contact = getContactByItem( contactListView_->currentItem() );
  if ( contact == 0 )
  {
    kdDebug() << "KMessView - moveCurrentContactToGroup() - Couldn't find the contact." << endl;
    return;
  }
  // Ask the server to remove the contact from the current group and add it to the given group.
  emit requestRemoveContact( contact->getHandle(), "FL", contact->getMetaGroupId() );
  emit    requestAddContact( contact->getHandle(), "FL", groupId                   );
}



// Move the current group down 1 in the contact list
void ContactList::moveCurrentGroupDown()
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - Move group down" << endl;
#endif
  Group *group, *nextGroup;
  int    sortPosition;

  group = getGroupByItem( contactListView_->currentItem() );
  if ( group != 0 )
  {
    // Get the group's sort position
    sortPosition = group->getSortPosition();
    // Get the group below this group
    nextGroup = getGroupBySortPosition( sortPosition + 1 );
    if ( nextGroup != 0 )
    {
      // Move the current group down
      group->setSortPosition( sortPosition + 1 );
      // Move the next group up
      nextGroup->setSortPosition( sortPosition );
      // Sort the contact list
      contactListView_->sort();
    }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
    else
    {
      kdDebug() << "ContactList - moveCurrentGroupDown - Couldn't find the group with sort position " << ( sortPosition + 1 ) << "." << endl;
    }
#endif
  }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  else
  {
    kdDebug() << "ContactList - moveCurrentGroupDown - Couldn't find the current group." << endl;
  }
#endif
}



// Move the current group up 1 in the contact list
void ContactList::moveCurrentGroupUp()
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - Move group up" << endl;
#endif
  Group *group, *previousGroup;
  int    sortPosition;

  group = getGroupByItem( contactListView_->currentItem() );
  if ( group != 0 )
  {
    // Get the group's sort position
    sortPosition = group->getSortPosition();
    // Get the group below this group
    previousGroup = getGroupBySortPosition( sortPosition - 1 );
    if ( previousGroup != 0 )
    {
      // Move the current group down
      group->setSortPosition( sortPosition - 1 );
      // Move the next group up
      previousGroup->setSortPosition( sortPosition );
      // Sort the contact list
      contactListView_->sort();
    }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
    else
    {
      kdDebug() << "ContactList - moveCurrentGroupUp - Couldn't find the group with sort position " << ( sortPosition - 1 ) << "." << endl;
    }
#endif
  }
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  else
  {
    kdDebug() << "ContactList - moveCurrentGroupUp - Couldn't find the current group." << endl;
  }
#endif
}



// Move the contact to the given group.
void ContactList::moveGivenContactToGroup( ContactListViewItem *contactMoving, ContactListViewItem *destinationGroup )
{
  const Contact *contact;
  const Group   *group;

  // Find the class that corresponds to the group item
  group   = getGroupByItem( destinationGroup );

  // Find the class that corresponds to the contact item
  contact = getContactByItem( contactMoving );

  if ( contact != 0 && group != 0 )
  {
    // Make sure the group isn't a special group
    if (    ( group->getId() != SpecialGroups::ALLOWED ) && ( group->getId() != SpecialGroups::REMOVED )
         && ( group->getId() != SpecialGroups::ONLINE  ) && ( group->getId() != SpecialGroups::OFFLINE ) )
    {
      // Ask the server to remove the contact from the current group and add it to the given group.
      // WARNING : NO ERROR CHECKING YET!
      emit requestRemoveContact( contact->getHandle(), "FL", contact->getMetaGroupId() );
      emit    requestAddContact( contact->getHandle(), "FL", group->getId()            );
    }
  }
}



// Bring up a dialog to get a group name, then add the group.
void ContactList::newGroupDialog()
{
  QString         newGroupName;
  bool             addGroup;
  GroupNameDialog *nameDialog;


  // Set a default group name
  newGroupName = "New Group";
  // Launch a dialog to get a new group name
  nameDialog = new GroupNameDialog(this);
  addGroup = nameDialog->launch(newGroupName);
  delete nameDialog;
  // Check that the user didn't cancel
  if ( addGroup )
  {
    // Request that the group be renamed
    emit requestAddGroup( newGroupName );
  }
}


// Read in account and other properties
void ContactList::readProperties(KConfig */*config*/)
{
}



// Remove a contact from a given list (and possibly group)
void ContactList::removeContact( QString handle, QString list, QString /*groupId*/ )
{
  Contact *contact;
  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - removeContact - Couldn't find contact " << handle << " in the contact list." << endl;
    return;
  }
  contact->setList( list, false );
  checkContactGroup( contact );
}



// Request that the current group be removed.
void ContactList::removeCurrentGroup()
{
  const Group     *group;
  QString         groupId;

  // Get the group item
  group = getGroupByItem( contactListView_->currentItem() );
  // If the group was found
  if ( group == 0 )
  {
    kdDebug() << "ContactList - removeCurrentGroup() - Couldn't find a matching group." << endl;
    return;
  }
  // If the group is not a special group...
  if (   ( group->getId() == SpecialGroups::ALLOWED )
      || ( group->getId() == SpecialGroups::REMOVED )
      || ( group->getId() == SpecialGroups::ONLINE  )
      || ( group->getId() == SpecialGroups::OFFLINE ) )
  {
    // Show a message that it can't be deleted
  	KMessageBox::error( 0, i18n("This is a special group and can't be changed.") );
    return;
  }
  // Ask the user for confirmation
  if ( ( KMessageBox::warningYesNoCancel( this, i18n("Are you sure you want to remove this group?")) ) == KMessageBox::Yes )
  {
    // signal a request that this group be removed
    emit requestRemoveGroup( group->getId() );
  }
}



// Remove a group
void ContactList::removeGroup( QString groupId )
{
  Group *group;
  group = getGroupById( groupId );
  if ( group == 0 )
  {
    kdDebug() << "ContactList - removeGroup - Group " << groupId << " doesn't exist in the contact list." << endl;
    return;
  }
#ifdef KMESSTEST
  uint groupCount, removedGroupCount;
  groupCount = groups_.count();
  removedGroupCount = removedGroups_.count();
#endif
  // Take the group from the list of groups
  groups_.remove( group );
  // ... and put it in the list of groups to be deleted when the the class is deleted
  removedGroups_.append( group );
  // Make the group invisible
  group->getItem()->setVisible( false );
#ifdef KMESSTEST
  ASSERT( groups_.count() == ( groupCount - 1 ) );
  ASSERT( removedGroups_.count() == ( removedGroupCount + 1 ) );
#endif
}



// Change a contact's name
void ContactList::renameContact( QString handle, QString newName )
{
  Contact *contact;
  contact = getContactByHandle( handle );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - renameContact - Couldn't find contact " << handle << " in the contact list." << endl;
    return;
  }
  contact->setFriendlyName( newName );
}



// Change a group's name
void ContactList::renameGroup( QString groupId, QString groupName )
{
  Group *group;
  group = getGroupById( groupId );
  if ( group == 0 )
  {
    kdDebug() << "ContactList - changeGroupName - Group " << groupId << " doesn't exist in the contact list." << endl;
    return;
  }
  group->setName( groupName );
}



// Launch a dialog to get a new group name, then request that the group
//  be renamed.
void ContactList::renameCurrentGroup()
{
  const Group     *group;
  QString         newGroupName;
  bool             changeName = false;
  GroupNameDialog *nameDialog;
  // Get the group item
  group = getGroupByItem( contactListView_->currentItem() );
  // If the group was found
  if ( group == 0 )
  {
    kdDebug() << "ContactList - removeCurrentGroup() - Couldn't find a matching group." << endl;
    return;
  }
  // If the group is not a special group...
  if (   ( group->getId() == SpecialGroups::ALLOWED )
      || ( group->getId() == SpecialGroups::REMOVED )
      || ( group->getId() == SpecialGroups::ONLINE  )
      || ( group->getId() == SpecialGroups::OFFLINE ) )
  {
    // Show a message that it can't be deleted
  	KMessageBox::error( 0, i18n("This is a special group and can't be changed.") );
    return;
  }
  // Otherwise, get the new group name..
  newGroupName = group->getName();
  // Launch a dialog to get a new group name
  nameDialog = new GroupNameDialog(this);
  changeName = nameDialog->launch(newGroupName);
  delete nameDialog;
  // Check that the user didn't cancel
  if ( changeName )
  {
    // Request that the group be renamed
    emit requestRenameGroup( group->getId(), newGroupName );
  }
  //contactListView_->currentItem()->startRename( 0 );
}



// Request that the current contact be added to the friends list.
void ContactList::requestAddContact()
{
  const Contact *contact;

  contact = getContactByItem( contactListView_->currentItem() );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - requestAddContact() - The current contact couldn't be found." << endl;
    return;
  }
  // Add a contact by unblocking it then adding it to the allowed list
  //  and the friends list (at group 0 by default - the user can move
  //  it later)
  emit requestAddContact   (contact->getHandle(), "FL", "0");
  emit requestRemoveContact(contact->getHandle(), "BL", "0");
  emit requestAddContact   (contact->getHandle(), "AL", "0");
}



// Allow the current contact to see the user's online status.
void ContactList::requestAllowContact()
{
  const Contact *contact;

  contact = getContactByItem( contactListView_->currentItem() );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - requestAllowContact() - The current contact couldn't be found." << endl;
    return;
  }
  // Add the contact to the allowed list
  emit requestAddContact(contact->getHandle(), "AL", "0");
}



// Request that the current contact be blocked.
void ContactList::requestBlockContact()
{
  QString contactHandle;

  contactHandle = getContactHandleByItem( contactListView_->currentItem() );

  if ( contactHandle != "" )
  {
    // Block a contact by removing it from the allowed list and adding
    //  it to the blocked list
    emit requestRemoveContact( contactHandle, "AL", "0" );
    emit requestAddContact   ( contactHandle, "BL", "0" );
  }
}



// Chat with the given contact
void ContactList::requestChatWithContact(QListViewItem *item)
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - Chat with contact" << endl;
#endif
  QString contactHandle = getContactHandleByItem( item );
  if ( contactHandle == "" )
  {
    kdDebug() << "ContactList - requestChatWithContact() - The contact couldn't be found." << endl;
    return;
  }
  // Signal a chat request
  emit chatWithContact( contactHandle );
}



// Request a chat with the current contact.
void ContactList::requestChatWithCurrentContact()
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - Chat with current contact" << endl;
#endif
  requestChatWithContact( contactListView_->currentItem() );
}



// Request that the current contact be removed from the friends list and blocked.
void ContactList::requestRemoveContact()
{
  const Contact *contact;

  contact = getContactByItem( contactListView_->currentItem() );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - requestRemoveContact() - The current contact couldn't be found." << endl;
    return;
  }
  // Remove a contact by blocking it, then removing it from the allowed
  //  list and the friends list.
  emit requestRemoveContact( contact->getHandle(), "AL", "0" );
  emit requestAddContact   ( contact->getHandle(), "BL", "0" );
  if ( contact->isFriend() )
  { // Check if it's on the friends list - it may just be "allowed"
    emit requestRemoveContact(contact->getHandle(), "FL", "");
  }
}



// Request that a contact's friendly name be changed to some new name.
void ContactList::requestRenameContact()
{
  const Contact     *contact;
  QString           newName;
  ContactNameDialog *nameDialog;
  bool               doChange;

  contact = getContactByItem( contactListView_->currentItem() );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - requestRenameContact() - The current contact couldn't be found." << endl;
    return;
  }

  //contactListView_->currentItem()->startRename( 0 );
  // Create a dialog to get the contact name
  nameDialog = new ContactNameDialog( this, "contactNameDialog" );
  // Get the current name as default
  newName = contact->getFriendlyName();
  // Show the dialog
  doChange = nameDialog->launch( newName );
  delete nameDialog;
  // If the user clicked "OK"
  if ( doChange )
  {
    // Add the contact to the allowed list
    emit requestRenameContact(contact->getHandle(), newName);
  }
}



// Request that the current contact be blocked.
void ContactList::requestUnblockContact()
{
  QString contactHandle;

  contactHandle = getContactHandleByItem( contactListView_->currentItem() );
  if ( contactHandle == "" )
  {
    kdDebug() << "ContactList - requestUnblockContact() - The current contact couldn't be found." << endl;
    return;
  }
  // Unblock a contact by removing it from the blocked list and adding
  //  it to the allowed list
  emit requestRemoveContact( contactHandle, "BL", "0" );
  emit requestAddContact   ( contactHandle, "AL", "0" );
}



// Save group properties
void ContactList::saveProperties(KConfig *config)
{
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
  kdDebug() << "ContactList - saving properties" << endl;
#endif
  Group *group;
  for ( group = groups_.first(); group; group = groups_.next() )
  {
#ifdef KMESSDEBUG_CONTACTLIST_GENERAL
    kdDebug() << "ContactList - saving group " << group->getName() << endl;
#endif
    group->saveProperties( config );
  }
}




// Pop up the html browser and show the msn member profile for
//  the current contact.
void ContactList::showContactMsnProfile()
{
  QString   contactHandle, urlPath;
  KURL      *url;
  KRun      *run;

  contactHandle = getContactHandleByItem( contactListView_->currentItem() );
  if ( contactHandle == "" )
  {
    kdDebug() << "ContactList - showContactMsnProfile() - The current contact couldn't be found." << endl;
    return;
  }
  // Set a path to the msn profile page.  Unfortunately, this isn't
  //  internationalized.
  urlPath = "http://members.msn.com/default.msnw?mem=" + contactHandle;
  // Create a URL to the given path
  url = new KURL( urlPath );
  // Launch the default html program for the given URL
	run = new KRun( *url );
}



// Show the contact properties dialog
void ContactList::showContactProperties()
{
  Contact                 *contact;
  ContactPropertiesDialog *dialog;

  contact = getContactByItem( contactListView_->currentItem() );
  if ( contact == 0 )
  {
    kdDebug() << "ContactList - showContactProperties() - The current contact couldn't be found." << endl;
    return;
  }
  dialog = new ContactPropertiesDialog( this, "contactProperties" );
  dialog->launch( contact );
  delete dialog;
}



// Check all contact groups when the view mode is updated
void ContactList::updateViewMode()
{
  Contact *contact;
  Group   *group;

  for ( group = groups_.first(); group; group = groups_.next() )
  {
    group->updateItemVisibility();
  }
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    contact->updateItemVisibility();
  }
  // Check the group membership of all contacts
  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    // Check the contact group, telling the function to not update the groups involved
    checkContactGroup( contact, false );
  }
  // Finally, update all group identifiers
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    group->updateItemIdentifier();
  }
}



// Update group and contact visibilities, like when "show allowed" or "show removed" is changed.
void ContactList::updateVisibilities()
{
  Contact *contact;
  Group   *group;

  for ( contact = contacts_.first(); contact; contact = contacts_.next() )
  {
    contact->updateItemVisibility();
  }
  for ( group = groups_.first(); group; group = groups_.next() )
  {
    group->updateItemVisibility();
    // The online and offline group identifiers depend on the visibilities of their children
    group->updateItemIdentifier();
  }
}
