/***************************************************************************
                          contactlist.h  -  description
                             -------------------
    begin                : Sun Jan 5 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CONTACTLIST_H
#define CONTACTLIST_H

#include <qptrlist.h>
#include <qwidget.h>

// Forward declarations
class Contact;
class ContactListView;
class ContactListViewItem;
class ContactListToolTip;
class CurrentAccount;
class Group;
class KAction;
class KActionMenu;
class KConfig;
class KMessTest;
class KPopupMenu;
class QListView;
class QListViewItem;

/**The list of contacts.
  *@author Mike K. Bennett
  */

class ContactList : public QWidget
{
  Q_OBJECT

  friend class KMessTest;

  public: // Public methods
    // The constructor
  	                     ContactList();
    // The destructor
  	                    ~ContactList();
    // Add a group to the contact list
    void                 addGroup( QString groupId, QString groupName );
    // Change a contact's status (and update the friendly name)
    void                 changeContactStatus( QString handle, QString friendlyName, QString status );
    // Set a contact to offline
    void                 contactOffline( QString handle );
    // Delete all the contacts
    void                 deleteAllContacts();
    // Delete all non-special groups
    void                 deleteNonSpecialGroups();
    // Return the contact with the given handle
    Contact*             getContactByHandle( QString handle );
    // Return the contact with the given list view item
    Contact*             getContactByItem( QListViewItem *item );
    // Get the friendly name of the given contact
    QString              getContactFriendlyNameByHandle( const QString& handle );
    // Return the list of contacts
    const QPtrList<Contact>& getContactList() const;
    // Get a pointer to the contact list dragged
    Contact*             getContactLastDragged();
    // Initialize the contact list
    bool                 initialize(ContactListView *contactListView);
    // Change a contact in the contact list or add it if it doesn't exist
    void                 modifyContact( QString handle, QString friendlyName, QString list, QString groupId );
    // Change a contact in the contact list or add it if it doesn't exist
    void                 modifyContact( QString handle, QString friendlyName, int list, QString groupId );
    // Read in account and other properties
    void                 readProperties(KConfig *config);
    // Remove a contact from a given list (and possibly group)
    void                 removeContact( QString handle, QString list, QString groupId );
    // Remove a group
    void                 removeGroup( QString groupId );
    // Change a contact's name
    void                 renameContact( QString handle, QString newName );
    // Change a group's name
    void                 renameGroup( QString groupId, QString newName );
    // Save account and other properties
    void                 saveProperties(KConfig *config);

  public slots: // Public slots
    // Bring up a dialog to get a group name, then add the group.
    void                 newGroupDialog();

  private: // Private methods
    // Add a contact to the contact list
    void                 addContact( QString handle, QString friendlyName, QString list, QString groupId);
    // Add a contact to the contact list
    void                 addContact( QString handle, QString friendlyName, int list, QString groupId);
    // Check that a contact is in the correct group
    void                 checkContactGroup( Contact *contact, bool updateGroups = true );
    // Return whether or not the contact with the given handle exists
    bool                 contactExists( QString handle );
    // Return the handle of the contact with the given list view item
    QString              getContactHandleByItem( QListViewItem *item );
    // Return the group with the given id
    Group*               getGroupById( QString groupId );
    // Return the group with the given list view item
    Group*               getGroupByItem( QListViewItem *item );
    // Return the group with the given sort value
    Group*               getGroupBySortPosition( int sortPosition );
    // Return whether or not a group with the given id exists
    bool                 groupExists( QString groupId );

  private slots: // Private slots
    // Collapse a group item
    void                 collapseGroup( QListViewItem *item );
    // A contact has added this user to their contact list.  Show a dialog
    //   on whether or not to add this contact.
    void                 contactAddedUser(QString handle, QString friendlyName);
    // Using hotmail's "compose", send an email to the current contact.
    void                 emailCurrentContact();
    // Expand a group item
    void                 expandGroup( QListViewItem *item );
    // Forward from a contact that it is offline
    void                 forwardContactOffline(const Contact *contact);
    // Forward from a contact that it is contact is online
    void                 forwardContactOnline(const Contact *contact);
    // Initialize the menu that pops up when a contact is right-clicked.
    void                 initContactActionPopup();
    // Initialize the menu that pops up when a group is right-clicked.
    void                 initGroupActionPopup();
    // An item in the contact list was renamed using in-place renaming.  We must decide what to do...
    void                 itemRenamed(QListViewItem *item, int column, const QString& text);
    // An item in the contact list was right-clicked, so bring up a menu
    void                 itemRightClicked(QListViewItem *item, const QPoint& point, int);
    // Move the current contact to the given group.
    void                 moveCurrentContactToGroup(QString groupId);
    // Move the current group down 1 in the contact list
    void                 moveCurrentGroupDown();
    // Move the current group up 1 in the contact list
    void                 moveCurrentGroupUp();
    // Move the contact to the given group.
    void                 moveGivenContactToGroup( ContactListViewItem *contactMoving, ContactListViewItem *destinationGroup );
    // Request that the current group be removed.
    void                 removeCurrentGroup();
    // Launch a dialog to get a new group name, then request that the group
    //  be renamed.
    void                 renameCurrentGroup();
    // Request that the current contact be added to the friends list.
    void                 requestAddContact();
    // Allow the current contact to see the user's online status.
    void                 requestAllowContact();
    // Request that the current contact be blocked.
    void                 requestBlockContact();
    // Chat with the given contact
    void                 requestChatWithContact(QListViewItem *item);
    // Request a chat with the current contact
    void                 requestChatWithCurrentContact();
    // Request that the current contact be removed from the friends list
    //  and blocked.
    void                 requestRemoveContact();
    // Request that a contact's friendly name be changed to some new name.
    void                 requestRenameContact();
    // Request that the current contact be blocked.
    void                 requestUnblockContact();
    // Pop up the html browser and show the msn member profile for the
    //  given contact.
    void                 showContactMsnProfile();
    // Show the contact properties dialog
    void                 showContactProperties();
    // Check all contact groups when the view mode is updated
    void                 updateViewMode();
    // Update group and contact visibilities, like when "show allowed" or "show removed" is changed.
    void                 updateVisibilities();

  private: // Private attributes
    // KActions used in the contact popup menu
    KAction             *addContact_, *allowContact_, *blockContact_,
                        *unblockContact_, *chatWithContact_, *msnProfile_,
                        *removeContact_, *emailContact_, *contactProperties_;
    // KActions used in the group popup menu
    KAction             *addGroup_, *moveGroupDown_, *moveGroupUp_, *removeGroup_, *renameGroup_;
    // The menu of actions possible to perform on a given contact
    KPopupMenu          *contactActionPopup_;
    // The menu of actions possible to perform on a given group
    KPopupMenu          *groupActionPopup_;
    // KSelectActions used in the contact popup menu
    KActionMenu         *moveContactToGroup_;
    // The contact list view
    ContactListView     *contactListView_;
    // The list of contacts
    QPtrList<Contact>    contacts_;
    // A pointer to the instance of a current account
    CurrentAccount      *currentAccount_;
    // The list of groups
    QPtrList<Group>      groups_;
    // Whether or not the contact list has been initialized
    bool                 initialized_;
    // A list of groups to be deleted when the class is deleted
    QPtrList<Group>      removedGroups_;
    // The tooltip manager for the list, must be explicitly delected
    ContactListToolTip * toolTip_;

  signals: // Signals
    // Request a chat with the given contact.
    void                 chatWithContact(QString handle);
    // Signal that a contact is offline
    void                 contactOffline( const Contact *contact );
    // Signal that a contact is online
    void                 contactOnline( const Contact *contact );
    // Request that the given contact be added to the given group (if applicable) and list.
    void                 requestAddContact(QString handle, QString list, QString groupID);
    // Request that the given group be added.
    void                 requestAddGroup(QString groupName);
    // Request that the given contact be removed from the given group (if applicable) and list.
    void                 requestRemoveContact(QString handle, QString list, QString groupID);
    // Request the the group with the given ID be removed.
    void                 requestRemoveGroup(QString groupId);
    // Request that the group with the given ID be renamed to the given name.
    void                 requestRenameContact(QString handle, QString newName);
    // Request that the group with the given ID be renamed to the given name.
    void                 requestRenameGroup(QString groupId, QString newName);
};

#endif
