/***************************************************************************
                          group.cpp  -  description
                             -------------------
    begin                : Sun Jan 5 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "group.h"

#include <qlistview.h>

#include <kaction.h>
#include <kapp.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kiconloader.h>
#include <kpopupmenu.h>

#include "contactlistviewitem.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"
#include "../specialgroups.h"

// The constructor
Group::Group(QString id, QString name, QListView *contactListView)
 : id_(id),
   name_(name),
   sortPosition_( id.toInt() )
{
#ifdef KMESSTEST
  bool goodNumber;
  id_.toInt( &goodNumber );
  ASSERT( goodNumber );
  ASSERT( contactListView != 0 );
#endif
  item_ = new ContactListViewItem( contactListView );
  item_->sortChildItems( SpecialColumns::SORT, true );
  item_->setDropEnabled( true );
  //item_->setRenameEnabled( 0, true );

  readProperties();

  action_ = new KAction();
  action_->setText( name_ );
  connect( action_, SIGNAL(       activated() ),
           this,    SLOT  ( actionActivated() ) );
  updateItemIdentifier();
  updateItemVisibility();
}



// The destructor
Group::~Group()
{
  delete action_;
#ifdef KMESSDEBUG_GROUP
  kdDebug() << "Group " << name_ << " (" << id_ << ") DESTROYED" << endl;
#endif
}



// The action was activated
void Group::actionActivated()
{
  emit actionSelected( id_ );
}



// Return the group ID
const QString& Group::getId() const
{
  return id_;
}



// Return the group item
ContactListViewItem* Group::getItem()
{
  return item_;
}



// Return the group name
const QString& Group::getName() const
{
  return name_;
}



// Return the sort value
const int& Group::getSortPosition() const
{
  return sortPosition_;
}



// Plug the group action into the given popup menu
void Group::plugAction(KActionMenu *popup)
{
  if (    ( id_ != SpecialGroups::ONLINE )
       && ( id_ != SpecialGroups::OFFLINE )
       && ( id_ != SpecialGroups::ALLOWED )
       && ( id_ != SpecialGroups::REMOVED ) )
  {
    popup->insert( action_ );
  }
}



// Read in group properties
void Group::readProperties()
{
  bool         isOpen;
  QString      entry, group;
  KConfig     *config = kapp->config();
  KIconLoader  load;

  if (    ( id_ != SpecialGroups::ONLINE )
       && ( id_ != SpecialGroups::OFFLINE )
       && ( id_ != SpecialGroups::ALLOWED )
       && ( id_ != SpecialGroups::REMOVED ) )
  {
    group = "Profile_" + CurrentAccount::instance()->getHandle();
  }
  else
  {
    group = "Profile_GENERAL";
  }

  config->setGroup( group );

  entry = "group_" + id_ + "_isexpanded";
  isOpen        = config->readBoolEntry( entry, true        );
  entry = "group_" + id_ + "_sortPosition";
  sortPosition_ = config->readNumEntry ( entry, id_.toInt() );
  item_->setOpen( isOpen );
  if ( isOpen )
  {
    item_->setIcon( "2downarrow" );
    //item_->setPixmap( 0, load.loadIcon( "2downarrow", KIcon::Small ) );
  }
  else
  {
    item_->setIcon( "2rightarrow" );
    //item_->setPixmap( 0, load.loadIcon( "2rightarrow", KIcon::Small ) );
  }
#ifdef KMESSDEBUG_GROUP
  kdDebug() << "Group " << name_ << " (" << id_ << ") has loaded sort position " << sortPosition_ << ".  It's ID is " << id_ << "."<< endl;
#endif
}



// Save group properties
void Group::saveProperties(KConfig *config)
{
  QString entry, group;

  if (    ( id_ != SpecialGroups::ONLINE )
       && ( id_ != SpecialGroups::OFFLINE )
       && ( id_ != SpecialGroups::ALLOWED )
       && ( id_ != SpecialGroups::REMOVED ) )
  {
    group = "Profile_" + CurrentAccount::instance()->getHandle();
  }
  else
  {
    group = "Profile_GENERAL";
  }
  config->setGroup( group );

  entry = "group_" + id_ + "_isexpanded";
  config->writeEntry( entry, item_->isOpen() );
  entry = "group_" + id_ + "_sortPosition";
  config->writeEntry( entry, sortPosition_ );
}



// Change the group name
void Group::setName(const QString& newName)
{
#ifdef KMESSTEST
  ASSERT( newName != "" );
  ASSERT( ( id_ != SpecialGroups::ONLINE ) && ( id_ != SpecialGroups::OFFLINE ) && ( id_ != SpecialGroups::ALLOWED ) && ( id_ != SpecialGroups::REMOVED ) );
#endif
  if ( ( name_ != newName ) && ( newName != "" ) )
  {
    name_ = newName;
    //item_->setText( 0, name_ );
    updateItemIdentifier();
  }
}



// Change the sort position
void Group::setSortPosition(const int& sortPosition)
{
  sortPosition_ = sortPosition;
  updateItemIdentifier();
}



// Unplug the group from the popup menu
void Group::unplugAction(KActionMenu *popup)
{
  popup->remove( action_ );
}



// Update the item identifier
void Group::updateItemIdentifier()
{
  int                  onlineCount = 0, totalCount = 0, visibleCount = 0;
  ContactListViewItem *childItem;
  QString              countIdentifier, identifier, sortString;
  // The group item's identifier is based on its name, the number of contacts it holds,
  //  and the type of group it is.
  // Go through the item's child items and count how many there are and how many are online
  childItem = (ContactListViewItem *)item_->firstChild();
  while ( childItem )
  {
    // If the item is online, add it to the list of contacts online
    if ( childItem->getOnlineStatus() )
    {
      onlineCount++;
    }
    // If the item is visible, add it to that count
    if ( childItem->isVisible() )
    {
      visibleCount++;
    }
    // Increase the total number of contacts, regardless.
    totalCount++;
    // Get the next child
    childItem = (ContactListViewItem *)childItem->nextSibling();
  }
  // If the group is one of the special "online" or "offline" groups, the count identifier is
  //  the number of visible children
  if ( ( getId() == SpecialGroups::ONLINE ) || ( getId() == SpecialGroups::OFFLINE ) )
  {
    countIdentifier.sprintf( "%d", visibleCount );
  }
  else if ( ( getId() == SpecialGroups::ALLOWED ) || ( getId() == SpecialGroups::REMOVED ) )
  {
    // If the group is "allowed" or "removed", then it's the total number of children,
    //  though, really, they should all be visible.
    countIdentifier.sprintf( "%d", totalCount );
  }
  else
  {
    // Other groups show the number of online contacts over the total number of contacts
    countIdentifier.sprintf( "%d/%d", onlineCount, totalCount );
  }
  // Set the identifier as the group name plus the count identifier
  item_->setIdentifier( getName() );
  item_->setStatusIdentifier( countIdentifier );
  // Set the group's sort position
  //sortString.sprintf( "%d", sortPosition_ );
  //item_->setText( SpecialColumns::SORT, sortString );
  item_->setSortIndex( sortPosition_ );
  // Sort the contacts in the group
  item_->sort();
}



// Update the visibilitiy of the item
void Group::updateItemVisibility()
{
  CurrentAccount  *currentAccount;
  bool             visible;

  // Visibility depends on the view mode whether or not allowed and removed contacts are shown.
  currentAccount = CurrentAccount::instance();
  if ( currentAccount == 0 )
  {
    kdDebug() << "Contact - updateVisibility() - Couldn't get an instance of the current contact." << endl;
    return;
  }
  // If the group is the offline group and offline contacts aren't shown..
  if ( ( getId() == SpecialGroups::OFFLINE ) && ( !currentAccount->getShowOfflineContacts() ) )
  {
    visible = false;
  }
  else
  {
    if ( currentAccount->getShowContactsByGroup() )
    {
      // If the view mode is by group, the online and offline groups should be invisible
      if ( ( getId() == SpecialGroups::ONLINE ) || ( getId() == SpecialGroups::OFFLINE ) )
      {
        visible = false;
      }
      else if ( getId() == SpecialGroups::ALLOWED )
      {
        // Whether or not the allowed group is visible depends on the user's settings
        visible = currentAccount->getShowAllowedContacts();
      }
      else if ( getId() == SpecialGroups::REMOVED )
      {
        // Whether or not the removed group is visible depends on the user's settings
        visible = currentAccount->getShowRemovedContacts();
      }
      else
      {
        // Otherwise, the group is just a normal group and should be visible
        visible = true;
      }
    }
    else
    {
      // If the view mode is to show contacts by online/offline mode, only the online and offline groups should be visible
      if ( ( getId() == SpecialGroups::ONLINE ) || ( getId() == SpecialGroups::OFFLINE ) )
      {
        visible = true;
      }
      else
      {
        visible = false;
      }
    }
  }
  item_->setVisible( visible );
}

