/***************************************************************************
                          kmessview.cpp  -  description
                             -------------------
    begin                : Thu Jan 9 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "kmessview.h"

#include <qheader.h>
#include <qlabel.h>
#include <qlistview.h>

#include <kdebug.h>
#include <kiconloader.h>
#include <klocale.h>
#include <kurllabel.h>

#include "contact/contactlistview.h"
#include "currentaccount.h"
#include "kmessdebug.h"

// The constructor
KMessView::KMessView(QWidget *parent, const char *name )
  : KMessViewInterface(parent,name),
    initialized_(false)
{
}



// The destructor
KMessView::~KMessView()
{
#ifdef KMESSDEBUG_KMESSVIEW
  kdDebug() << "DESTROYED KMessView " << endl;
#endif
}



// The program is connected to the server.  Store the user account and
//  update the user information.
void KMessView::connected()
{
#ifdef KMESSDEBUG_KMESSVIEW
  kdDebug() << "KMessView - connected" << endl;
#endif
  // Run the update functions with the current user data
  updateEmailDisplay();
  updateUserStatus();
  updateNoEmails();
  // Enable the labels
  onlineStatusLabel_->setEnabled(true);
  emailLabel_->setEnabled(true);
}



// Called when the program has disconnected from the server.
void KMessView::disconnected()
{
  KIconLoader load;

  // Disable and reset the labels
  onlineStatusLabel_->setText( i18n("Disconnected") );
  onlineStatusLabel_->setEnabled(false);
  onlineStatusPixmapLabel_->setPixmap( load.loadIcon( "offline", KIcon::User ) );
  emailLabel_->setText("");
  emailLabel_->setEnabled(false);
}



// The email label was clicked so open the user's preferred email url
void KMessView::emailLabelClicked()
{
  currentAccount_->openHotmailAtInbox();
}



// Initialize the class
bool KMessView::initialize()
{
  bool         success;
  KIconLoader  load;

  if ( initialized_ )
  {
    kdDebug() << "KMessView already initialized." << endl;
    return false;
  }
  success = initContactListView();
  if ( !success )
  {
    kdDebug() << "KMessView - Couldn't initialize contact list view." << endl;
    return false;
  }
  // Set an icon for the email label
  emailPixmapLabel_->setPixmap( load.loadIcon("email", KIcon::User) );
  // Get and connect and instance of the current account
  currentAccount_ = CurrentAccount::instance();
  connect( currentAccount_, SIGNAL(             changedNoEmails() ),
           this,            SLOT  (              updateNoEmails() ) );
  connect( currentAccount_, SIGNAL(               changedStatus() ),
           this,            SLOT  (            updateUserStatus() ) );
  connect( currentAccount_, SIGNAL(         changedFriendlyName() ),
           this,            SLOT  (            updateUserStatus() ) );
  connect( currentAccount_, SIGNAL( changedEmailDisplaySettings() ),
           this,            SLOT  (          updateEmailDisplay() ) );

  return true;
}



// Initialize the contact list view
bool KMessView::initContactListView()
{
  KIconLoader load;
  QPixmap     pixmap;

  // Hide the column header
	contactListView_->header()->hide();
  // Add hidden columns for group and sorting info
  contactListView_->addColumn("Sort", 0);
  contactListView_->addColumn("Status", 0);
  contactListView_->addColumn("MetaGroupId", 0);

  // Disable the horizontal scroll bar due to the asian character line-breaking thing
//  contactListView_->setHScrollBarMode( QScrollView::AlwaysOff );

  contactListView_->setPaletteForegroundColor( QColor( "#000000" ) );
  pixmap = load.loadIcon( "background", KIcon::User, 0, KIcon::DefaultState, 0L, true );
#ifdef KMESSDEBUG_KMESSVIEW
  kdDebug() << "KMessView - Loaded an icon for the list view background.  is null = " << pixmap.isNull() << "." << endl;
#endif
  if ( !pixmap.isNull() )
  {
#ifdef KMESSDEBUG_KMESSVIEW
    kdDebug() << "KMessView - Setting the background." << endl;
#endif

    contactListView_->setBackground(pixmap);
  }
  return true;
}



// Change whether or not the email label is displayed based on account settings.
void KMessView::updateEmailDisplay()
{
  if ( currentAccount_->getShowEmail() )
  {
    emailLabel_->setEnabled( true );
    emailPixmapLabel_->setEnabled( true );
    emailLabel_->setMaximumHeight( 700 );
    emailLabel_->setMaximumWidth( 700 );
    emailPixmapLabel_->setMaximumHeight( 16 );
    emailPixmapLabel_->setMaximumWidth( 16 );
  }
  else
  {
    emailLabel_->setEnabled( false );
    emailPixmapLabel_->setEnabled( false );
    emailLabel_->setMaximumHeight( 0 );
    emailLabel_->setMaximumWidth( 0 );
    emailPixmapLabel_->setMaximumHeight( 0 );
    emailPixmapLabel_->setMaximumWidth( 0 );
  }
}



// Update the email count.
void KMessView::updateNoEmails()
{
  QString emailString, message;

  // Put the number of emails in a string
  emailString.sprintf(" %d ", CurrentAccount::instance()->getNoEmails() );
  // Update the label
  if ( CurrentAccount::instance()->getNoEmails() == 1 )
  {
    message = i18n("You have") + emailString + i18n("new email in your inbox");
  }
  else
  {
    message = i18n("You have") + emailString + i18n("new emails in your inbox");
  }
  emailLabel_->setText( message );
  // Set the user's preferred email website.
  emailLabel_->setURL( CurrentAccount::instance()->getEmailUrl() );
}



// Update the user's status.
void KMessView::updateUserStatus()
{
  KIconLoader      load;
  QString          friendlyName, status;
  QString          iconName, statusIdentifier;

  friendlyName = CurrentAccount::instance()->getFriendlyName();
  status = CurrentAccount::instance()->getStatus();

#ifdef KMESSTEST_KMESSVIEW
  kdDebug() << "KMessView - update user status to " << status << endl;
#endif
  // Update the onlineStatusLabel and pixmap
  if ( status == "AWY" )
  {
    statusIdentifier = i18n( "(Away)" );
    iconName = "away";
  }
  else if ( status == "BRB" )
  {
    statusIdentifier = i18n( "(Be Right Back)" );
    iconName = "berightback";
  }
  else if ( status == "BSY" )
  {
    statusIdentifier = i18n( "(Busy)" );
    iconName = "busy";
  }
  else if ( status == "FLN" )
  {
    statusIdentifier = "";
    iconName = "offline";
  }
  else if ( status == "HDN" )
  {
    statusIdentifier = i18n("(Invisible)");
    iconName = "invisible";
  }
  else if ( status == "IDL" )
  {
    statusIdentifier = i18n( "(Away - Idle)" );
    iconName = "away";
  }
  else if ( status == "LUN" )
  {
    statusIdentifier = i18n( "(Out to Lunch)" );
    iconName = "lunch";
  }
  else if ( status == "NLN" )
  {
    statusIdentifier = "";
    iconName = "online";
  }
  else if ( status == "PHN" )
  {
    statusIdentifier = i18n( "(On the Phone)" );
    iconName = "onthephone";
  }
  else
  {
    kdDebug() << "KMessView - updateUserStatus() - The account had the invalid status " << status << "." << endl;
    return;
  }
  onlineStatusLabel_->setText( friendlyName + "  " + statusIdentifier );
  onlineStatusPixmapLabel_->setPixmap( load.loadIcon( iconName, KIcon::User ) );
}
