/***************************************************************************
                          transferfile.cpp  -  description
                             -------------------
    begin                : Sat Mar 22 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "transferfile.h"

#include <qfile.h>

#include <kdebug.h>
#include <kextsock.h>
#include <klocale.h>

#include "../../kmessdebug.h"
#include "../mimemessage.h"
#include "transferprogressdialog.h"

// The constructor
TransferFile::TransferFile(const QString& handle, const QString& localIp)
 : Application(localIp),
   bytesRemaining_(0),
   file_(0),
   socket_(0),
   handle_(handle),
   progressDialog_(0)
{
}



// The destructor
TransferFile::~TransferFile()
{
  deleteProgressDialog();
}



// Convert a char to an integer
int TransferFile::charToInt(char c)
{
  int i;
  i = (int)c;
  if ( i < 0 )
  {
    i += 256;
  }
#ifdef KMESSTEST
  ASSERT( ( i >= 0 ) && ( i < 256 ) );
#endif
  return i;
}



// Delete the progress dialog.
void TransferFile::deleteProgressDialog()
{
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "Deleting progress dialog." << endl;
#endif
  if ( progressDialog_ != 0 )
  {
    progressDialog_->close();
    delete progressDialog_;
    progressDialog_ = 0;
  }
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "Progress dialog deleted." << endl;
#endif
}



// Disconnect the socket.  Close the file.
void TransferFile::disconnect()
{
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Disconnect" << endl;
#endif
  // Delete the progress dialog
  deleteProgressDialog();
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Close the file" << endl;
#endif
  // Close the file before disconnecting
  if ( file_ != 0 )
  {
   	file_->flush();
    file_->close();
    // Clean up the pointers
    delete file_;
    file_ = 0;
  }
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Close the socket" << endl;
#endif
  // Close and delete the socket
  if ( socket_ != 0 )
  {
    socket_->flush();
    socket_->closeNow();
    delete socket_;
    socket_ = 0;
  }
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Disconnected at file transfer" << endl;
#endif
}



// Cancel the file transfer.
void TransferFile::failTransfer(QString text)
{
  QString html;

#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Fail transfer - " << text << endl;
#endif
  // Create a cancel message for the user.
  html  = "<font color=purple>";
  html += i18n("The transfer of %1 ").arg( "<font color=red>" + fileName_ + "</font>" );
  html += text;
  html += ".</font>";

  // Show the user the message
  emit messageToUser( html );

  // Disconnect
  disconnect();

#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Ask to be removed" << endl;
#endif
  // Signal that this receive transfer should be deleted.
  emit deleteMe( getCookie() );
}



// Return the application's GUID
QString TransferFile::getAppId()
{
  return "{5D3E02AB-6190-11d3-BBBB-00C04F795683}";
}



// Return the user's handle
const QString& TransferFile::getHandle() const
{
  return handle_;
}



// Create and initilize the progress dialog.
void TransferFile::initializeProgressDialog(QString description)
{
  if ( progressDialog_ == 0 )
  {
    // Create the dialog
    progressDialog_ = new TransferProgressDialog( description, fileSize_ );
    // Connect the dialog so that if the user closes it, it's deleted.
    connect( progressDialog_, SIGNAL( cancelTransfer() ),
             this,            SLOT  ( cancelTransfer() ) );
    // Show the dialog.
    progressDialog_->show();
  }
}



// Convert an integer (should be less than 256) to a char
char TransferFile::intToChar(int i)
{
#ifdef KMESSTEST
  ASSERT( ( i >= 0 ) && ( i < 256 ) );
#endif
  return (char)i;
}



// This is called when there's an error in the socket.
void TransferFile::socketError(int error)
{
  kdDebug() << "TransferFile - Socket error " << error << endl;
}



// Signal the user that the tranfer was successful and delete this object.
void TransferFile::successfulTransfer()
{
  QString fileURL, html;

#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Successful Transfer" << endl;
#endif

  if ( progressDialog_ != 0 )
  {
    progressDialog_->finishTransfer();
  }

  fileURL  = "<a href=\"file:/" + filePath_ + "\">";
  fileURL += fileName_ + "</a>";
  // Create a cancel message for the user.
  html  = "<font color=purple>";
  html += i18n("Successfully transferred file: %1").arg( "<font color=blue>" + fileURL + "</font>" );
  html += "</font>";

  // Show the user the message
  emit messageToUser( html );
  // Disconnect
  disconnect();
#ifdef KMESSDEBUG_TRANSFERFILE
  kdDebug() << "TransferFile - Ask to be removed" << endl;
#endif
  // Signal that this receive transfer should be deleted.
  emit deleteMe( getCookie() );
}



// Write a message to the socket.
void TransferFile::writeMessage(QString message)
{
  unsigned int bytesWritten;

  bytesWritten = socket_->writeBlock( message.utf8().data(), message.utf8().length() );
  if ( bytesWritten != message.length() )
  {
#ifdef KMESSDEBUG_TRANSFERFILE
    kdDebug() << "Tried to write " << message.length() << " bytes.  Wrote " << bytesWritten << " bytes." << endl;
#endif
  }
}



// Update the progress dialog with the number of bytes transferred.
void TransferFile::updateProgressDialog(int noBytes)
{
  if ( progressDialog_ != 0 )
  {
    progressDialog_->updateProgress( noBytes );
  }
}
