/***************************************************************************
                          msnnotificationconnection.cpp  -  description
                             -------------------
    begin                : Thu Jan 23 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "msnnotificationconnection.h"

#include <kconfig.h>
#include <kdebug.h>
#include <klocale.h>
#include <kmdcodec.h>
#include <kmessagebox.h>
#include <kpassdlg.h>
#include <krun.h>
#include <kurl.h>

#include "../contact/contact.h"
#include "../contact/contactlist.h"
#include "../contact/contactlistview.h"
#include "../currentaccount.h"
#include "../kmessdebug.h"
#include "chatinformation.h"
#include "mimemessage.h"
#include "sslloginhandler.h"

#ifdef KMESSDEBUG_NOTIFICATION
//  #define KMESSDEBUG_NOTIFICATION_GENERAL
  #define KMESSDEBUG_NOTIFICATION_MESSAGES
  #define KMESSDEBUG_NOTIFICATION_INVISIBLE
  #define KMESSDEBUG_NOTIFICATION_MSN9
#endif

// The constructor
MsnNotificationConnection::MsnNotificationConnection()
 : MsnConnection("NOTIFICATION"),
   initialized_(false),
   noContacts_(0),
   sslLoginHandler_(0),
   totalNoContacts_(0)
{
  // Connect the login timer to the checkLogin slot
  connect( &loginTimer_, SIGNAL(    timeout() ),
           this,         SLOT  ( checkLogin() ) );
}



// The destructor
MsnNotificationConnection::~MsnNotificationConnection()
{
  delete contactList_;
  delete sslLoginHandler_;
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "DESTROYED Notification" << endl;
#endif
}



// Add a contact to the given list or, if applicable, group
void MsnNotificationConnection::addContact(QString handle, QString list, QString groupId)
{
#ifdef KMESSTEST
  bool goodGroupId;
  groupId.toInt( &goodGroupId );
  ASSERT( ( goodGroupId ) || ( list != "FL" ) );
  ASSERT( ( list == "FL" ) || ( list == "AL" ) || ( list == "BL" ) );
#endif

  QString command;
  command = list + " " + handle + " " + handle;
  if ( list == "FL" )
  {
    command += " " + groupId;
  }
  command += "\r\n";
  sendCommand( "ADD", command );
}



// Add a new group
void MsnNotificationConnection::addGroup(QString name)
{
  addPercents( name );
  sendCommand( "ADG", name  + "\r\n" );
}



// Change a contact or the user's friendly name
void MsnNotificationConnection::changeFriendlyName( QString handle, QString newName )
{
  QString name = newName;
  addPercents( name );
  sendCommand( "REA", handle + " " + name + "\r\n" );
}



// Request a change in the user's status
void MsnNotificationConnection::changeStatus( const QString& newStatus )
{
  sendCommand( "CHG", newStatus + "\r\n" );
}



// Check whether login was successful within time limit
void MsnNotificationConnection::checkLogin()
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Login not successful within time limit." << endl;
#endif
  QString message;
  // Show a message to the user
  message = i18n("Connection timed out") + "\r\n";
  message += i18n("There may be a problem with the servers");
  KMessageBox::error( 0, message );
  closeConnection();
}



// Close the connection with the server
void MsnNotificationConnection::closeConnection()
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Close the connection." << endl;
#endif
  // Stop the login timer in case the user closed the connection during login
  loginTimer_.stop();
  noContacts_ = 0;
  totalNoContacts_ = 0;
  // Disconnect from the server
  disconnectFromServer();
  // Remove contacts and non-special groups from the contact list
  if ( contactList_ != 0 )
  {
    contactList_->deleteAllContacts();
    contactList_->deleteNonSpecialGroups();
  }
  if ( currentAccount_ != 0 )
  {
    currentAccount_->setNoEmails( 0 );
  }
}



// The socket connected, so send the version command.
void MsnNotificationConnection::connectionSuccess()
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Connected Asynchronously, so send VER." << endl;
#endif

  putVer();
}



// Create the SSL login handler
SslLoginHandler* MsnNotificationConnection::createLoginHandler()
{
  SslLoginHandler *sslLoginHandler;
  // Create the SSL login handler
  sslLoginHandler = new SslLoginHandler();
  connect( sslLoginHandler, SIGNAL(       loginFailed()        ),
           this,            SLOT  (    sslLoginFailed()        ) );
  connect( sslLoginHandler, SIGNAL(    loginSucceeded(QString) ),
           this,            SLOT  ( sslLoginSucceeded(QString) ) );
  return sslLoginHandler;
}



// Show a .net status message.
void MsnNotificationConnection::dotNetMessage(QString message)
{
  int    result;
  KURL   netURL( "http://messenger.msn.com/support/status.asp" );

  message += "\n" + i18n("Show .NET service status?");
  result = KMessageBox::questionYesNo( 0, message, i18n("Server Error") );
  if ( result == KMessageBox::Yes )
  {
	  new KRun( netURL );
  }
  closeConnection();
}



// Forward from the contact list that a contact is offline
void MsnNotificationConnection::forwardContactOffline(const Contact *contact)
{
#ifdef KMESSTEST
  ASSERT( contact != 0 );
#endif
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  if ( contact != 0 )
  {
    kdDebug() << "MsnNotificationConnection - Forwarding that " << contact->getFriendlyName() << " went offline." << endl;
  }
#endif
  emit contactOffline( contact );
}



// Forward from the contact list that a contact is online
void MsnNotificationConnection::forwardContactOnline(const Contact *contact)
{
#ifdef KMESSTEST
  ASSERT( contact != 0 );
#endif
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  if ( contact != 0 )
  {
    kdDebug() << "MsnNotificationConnection - Forwarding that " << contact->getFriendlyName() << " went online." << endl;
  }
#endif
  emit contactOnline( contact );
}



// Go online
void MsnNotificationConnection::goOnline()
{
//  kdDebug() << "Local IP is " << getLocalIp() << endl;
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Lists received.  Go online." << endl;
#endif
  int      ack;
  QString ackString;

  if ( currentAccount_ != 0 )
  {
#ifdef KMESSDEBUG_NOTIFICATION_INVISIBLE
    kdDebug() << "Notification: Prepare to change status to online or invisible." << endl;
    kdDebug() << "Notification: 'Start invisible' is " << currentAccount_->getStartInvisible() << "." << endl;
#endif
    // Change status to either online or invisible
    if ( currentAccount_->getStartInvisible() )
    {
#ifdef KMESSDEBUG_NOTIFICATION_INVISIBLE
      kdDebug() << "Notification: Change status to invisible." << endl;
#endif
      // Change status to invisible
      sendCommand("CHG", "HDN\r\n");
    }
    else
    {
#ifdef KMESSDEBUG_NOTIFICATION_INVISIBLE
      kdDebug() << "Notification: Change status to online." << endl;
#endif
      // Change status to online
      sendCommand("CHG", "NLN\r\n");
    }
    // Ask for inbox and compose URLs
    ack = sendCommand("URL", "INBOX\r\n");
    ackString.sprintf( "%d", ack );
    currentAccount_->setInboxInformation( ackString, "", "" );

    ack = sendCommand("URL", "COMPOSE\r\n");
    ackString.sprintf( "%d", ack );
    currentAccount_->setComposeInformation( ackString, "", "" );
  }
  else
  {
    kdDebug() << "Notification::goOnline() - WARNING - currentAccount_ is null!" << endl;
  }

  setSendPings( true );

  // Notify observers that the server is connected
  loginTimer_.stop();
  emit connected();
}



// Received the addition of a contact to a list or list and group
void MsnNotificationConnection::gotAdd(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotAdd - WARNING - Contact list is null!" << endl;
    return;
  }
  // Extract the handle, friendlyName, and list from the message
  QString friendlyName, groupId;

  friendlyName = command[5];
  removePercents( friendlyName );

  if ( command[2] == "FL" )
  {
    groupId    = command[6];
  }
  else
  {
    groupId = QString::null;
  }

#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Added " << friendlyName << " (" << command[4] << ") to " << command[2] << " " << groupId << "." << endl;
#endif
  // Give the information to the contact list.
  contactList_->modifyContact( command[4], friendlyName, command[2], groupId );
}



// Received the addition of a group
void MsnNotificationConnection::gotAdg(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotAdg - WARNING - Contact list is null!" << endl;
    return;
  }
  QString name = command[3];
  removePercents( name );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Added group " << name << " (" << command[4] << ")." << endl;
#endif
  // Pass the information to the group list.
  contactList_->addGroup( command[4], name );
}



// Received confirmation of the user's status change
void MsnNotificationConnection::gotChg(const QStringList& command)
{
  if ( currentAccount_ != 0 )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification: User's status changed to " << command[2] << "." << endl;
#endif
    // Change the user's status
    currentAccount_->setStatus( command[2] );
  }
  else
  {
    kdDebug() << "Notification::gotChg() - WARNING - currentAccount_ is null!" << endl;
  }
}



// Received a challenge from the server
void MsnNotificationConnection::gotChl(const QStringList& command)
{
  // Get the challengeCode from the message
  QString addCode = "Q1P7W2E4J9R8U3S5";

  // Get the md5 context of the challenge code combined with the stored code
  KMD5 context( command[2] + addCode );

  sendCommand("QRY", "msmsgs@msnmsgr.com 32\r\n" + context.hexDigest() );
}



// Received a version update from the server
void MsnNotificationConnection::gotCvr(const QStringList& /*command*/)
{
  if ( currentAccount_ != 0 )
  {
    // Send the USR command
    sendCommand( "USR", "TWN I " + currentAccount_->getHandle() + "\r\n" );
  }
}



// Received notice that a contact went offline
void MsnNotificationConnection::gotFln(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotFln - WARNING - Contact list is null!" << endl;
    return;
  }
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: " << command[1] << " went offline." << endl;
#endif
  contactList_->contactOffline( command[1] );
}



// Received notice that a contact is already online
void MsnNotificationConnection::gotIln(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotIln - WARNING - Contact list is null!" << endl;
    return;
  }
  // Get the contact info from the message
  QString friendlyName = command[4];
  removePercents( friendlyName );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: " << friendlyName << " (" << command[3] << ") is initially " << command[2] << "." << endl;
#endif
  contactList_->changeContactStatus( command[3], friendlyName, command[2] );

  QString msnObject = command[6];
#ifdef KMESSDEBUG_NOTIFICATION_MSN9
  if ( !msnObject.isEmpty() )
  {
    removePercents( msnObject );
    kdDebug() << "Notification: THIS PERSON IS USING MSN6. Obj: " << msnObject << endl;
  }
#endif
}



// Received authentication protocols
void MsnNotificationConnection::gotInf(const QStringList& command)
{
#ifdef KMESSTEST
  ASSERT( command[2] == "MD5" );
#endif
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Got authentication protocols." << endl;
#endif
  if ( currentAccount_ != 0 )
  {
    // Send user information
    sendCommand( "USR", "MD5 I " + currentAccount_->getHandle() + "\r\n" );
  }
  else
  {
    kdDebug() << "Notification::gotInf() - WARNING - currentAccount_ is null!" << endl;
  }
}



// Received group information
void MsnNotificationConnection::gotLsg(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotLsg - WARNING - Contact list is null!" << endl;
    return;
  }
  QString name = command[2];
  removePercents( name );
#ifdef KMESSDEBUG_NOTIFICATION_MSN9
  kdDebug() << "Notification: Got group " << name << "(" << command[1] << ")" << endl;
#endif
#ifdef KMESSTEST
  bool goodId;
  command[1].toInt( &goodId );
  ASSERT( goodId );
#endif
  contactList_->addGroup( command[1], name );
}



// Received contact information
void MsnNotificationConnection::gotLst(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotLst - WARNING - Contact list is null!" << endl;
    return;
  }
  //emit statusMessage( i18n("Getting lists"), 2 );
#ifdef KMESSDEBUG_NOTIFICATION_MSN9
  kdDebug() << "Notification: Got LST : " << command.join( " " ) << endl;
#endif
  QString handle = command[1];
  QString friendlyName = command[2];
  removePercents( friendlyName );
  int group = command[3].toInt();
  QString groupId = command[4];
#ifdef KMESSDEBUG_NOTIFICATION_MSN9
  kdDebug() << "Notification: Got contact " << friendlyName << "(" << handle << ") " << group << " " << groupId << "." << endl;
#endif
  contactList_->modifyContact( handle, friendlyName, group, groupId );


  // Check if this is the last LST
  noContacts_++;
  if ( noContacts_ == totalNoContacts_ )
  {
#ifdef KMESSDEBUG_NOTIFICATION_MSN9
    kdDebug() << "Notification: All " << totalNoContacts_ << " were received.  Go online." << endl;
#endif
    goOnline();
  }
/*
  // Pull the contact's information from the list message.
  // Not all list messages are the same, only the FL messages
  //  contain group information, for instance.
  QString friendlyName, groupId;

  friendlyName = command[7];
  removePercents( friendlyName );

  if ( command[2] == "FL" )
  {
    groupId    = command[8];
#ifdef KMESSTEST
    bool goodId;
    groupId.toInt( &goodId );
    ASSERT( goodId );
#endif
  }
  else
  {
    groupId = QString::null;
  }

#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Got contact " << friendlyName << "(" << command[6] << ") " << command[2] << "." << endl;
#endif
  contactList_->modifyContact( command[6], friendlyName, command[2], groupId );

  // Check if the list entry received is the final list entry
  if ( command[2] == "RL" )
  { // When the reverse list is done, then all the contacts have been
    //  updated.  The reverse list is done when the contact number
    //  equals the total number.
   if ( command[4] == command[5] )
    { // Do what needs to be done when officially online.
      goOnline();
    }
  }
*/
}



// Received notice that a contact has changed status
void MsnNotificationConnection::gotNln(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotNln - WARNING - Contact list is null!" << endl;
    return;
  }
  // Get the contact info from the message
  QString friendlyName = command[3];
  removePercents( friendlyName );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: " << friendlyName << " (" << command[2] << ") is now " << command[1] << "." << endl;
  kdDebug() << "Notification: Rename the contact in the list?" << endl;
#endif
  // If the contact's name has changed, change it in MSN's version of the contact list.
  QString currentFriendlyName = contactList_->getContactFriendlyNameByHandle( command[2] );
  if ( ( friendlyName != currentFriendlyName ) && ( friendlyName != command[2] ) )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification: The current friendly name of " << currentFriendlyName << " doesn't match " << friendlyName << ".  Rename the contact in the list." << endl;
#endif
    changeFriendlyName( command[2], friendlyName );
  }
  contactList_->changeContactStatus( command[2], friendlyName, command[1] );
}



// Received notice of disconnetion from the server
void MsnNotificationConnection::gotOut(const QStringList& command)
{
  // If the first word is "OTH", it means that this user connected from elsewhere.
  if ( command[1] == "OTH" )
  {
  	KMessageBox::error( 0, i18n("You have been disconnected because you \nconnected from another MSN Messenger client \nor from another location") );
	}
  closeConnection();
}



// Received confirmation of the user's or a contact's renaming
void MsnNotificationConnection::gotRea(const QStringList& command)
{
  if ( currentAccount_ == 0 )
  {
    kdDebug() << "Notification::gotRea() - WARNING - currentAccount_ is null!" << endl;
    return;
  }
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotRea - WARNING - Contact list is null!" << endl;
    return;
  }
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: " << command[3] << " renamed to " << command[4] << "." << endl;
#endif
  // Check if this REA applies to the user or a contact
  if ( command[3] == currentAccount_->getHandle() )
  {
    QString friendlyName = command[4];
    removePercents( friendlyName );
    // Change the user's name
    currentAccount_->setFriendlyName( friendlyName );
  }
}



// Received confirmation of change in a group's name
void MsnNotificationConnection::gotReg(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotReg - WARNING - Contact list is null!" << endl;
    return;
  }
  QString name = command[4];
  removePercents( name );

#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Group " << command[3] << " renamed to " << name << "." << endl;
#endif
  // Pass the info to the group list
  contactList_->renameGroup( command[3], name );
}



// Received confirmation of a conact's removal from list or list and group
void MsnNotificationConnection::gotRem(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotRem - WARNING - Contact list is null!" << endl;
    return;
  }
  // Extract the handle, friendlyName, and list from the message
  QString groupId;
  if ( command[2] == "FL" )
  {
    groupId = command[5];
  }
  else
  {
    groupId = QString::null;
  }
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Contact " << command[4] << " removed from " << command[2] << " " << groupId << "." << endl;
#endif
  // Pass the information to the contact list
  contactList_->removeContact( command[4], command[2], groupId );
}



// Received confirmation of a group's removal
void MsnNotificationConnection::gotRmg(const QStringList& command)
{
  if ( contactList_ == 0 )
  {
    kdDebug() << "Notification::gotRmg - WARNING - Contact list is null!" << endl;
    return;
  }
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Remove group " << command[3] << "." << endl;
#endif
  // Pass the info to the group list
  contactList_->removeGroup( command[3] );
}



// Received a chat request from a contact
void MsnNotificationConnection::gotRng(const QStringList& command)
{
  // Pull the data from the message
  QString friendlyName  = command[6];
  removePercents( friendlyName );

  QString serverAndPort = command[2];
  QString server     = serverAndPort.left ( serverAndPort.find(":") );
  QString portString = serverAndPort.right( serverAndPort.length() - serverAndPort.find(":") - 1 );
  bool goodPort;
  int port = portString.toInt( &goodPort );
  if ( !goodPort )
  {
    kdDebug() << "Notification::gotRng - WARNING - Couldn't get port from string " << portString << endl;
    return;
  }
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: " << friendlyName << " (" << command[5] << ") requesting a chat at " << server << ":" << port << " id = " << command[1] << " auth = " << command[4] << "." << endl;
#endif
  // Create a chat information object
  ChatInformation *chatInfo = new ChatInformation( command[5], contactList_, server, port, command[4], command[1] );
  emit startChat( chatInfo );
}



// Received synchronization information
void MsnNotificationConnection::gotSyn(const QStringList& command)
{
  emit ( i18n("Got synchronization") );
  totalNoContacts_ = command[3].toInt();
#ifdef KMESSDEBUG_NOTIFICATION_MSN9
  kdDebug() << "Got SYN.  # of contacts is " << totalNoContacts_ << endl;
#endif
}



// Received the folder and command info for Hotmail's inbox or compose
void MsnNotificationConnection::gotUrl(const QStringList& command)
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Got URL information." << endl;
#endif
  if ( currentAccount_ != 0 )
  {
    // Check the inbox and compose URLs to see if they're storing this ack
    if ( currentAccount_->getInboxCommand() == command[1] )
    { // This is the inbox url
      currentAccount_->setInboxInformation( command[4], command[2], command[3] );
    }
    else if ( currentAccount_->getComposeCommand() == command[1] )
    { // This is the compose url
      currentAccount_->setComposeInformation( command[4], command[2], command[3] );
    }
  }
  else
  {
    kdDebug() << "Notification::gotUrl() - WARNING - currentAccount_ is null!" << endl;
  }
}



// Received user authentication information
void MsnNotificationConnection::gotUsr(const QStringList& command)
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Got USR information." << endl;
#endif
  if ( currentAccount_ != 0 )
  {
    // This is either asking for authentication or confirming it
    if ( command[2] == "TWN" )
    {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
      kdDebug() << "Notification: USR authentication." << endl;
#endif
      QString parameters;
      parameters = command[4];
      removePercents( parameters );
      // Change statusbar
      emit statusMessage( i18n("Authenticating"), 2 );
      // Create the sslLoginHandler
      sslLoginHandler_ = createLoginHandler();
      // Send the SSL login handler the parameter list.
      sslLoginHandler_->login( parameters, currentAccount_->getHandle(), currentAccount_->getPassword() );
/*
      // Create an md5 context
      QString hash;
      QString password = currentAccount_->getPassword();
      // Check if a password has been saved
      if ( password.isEmpty() )
      {
        // If not, show a dialog
        QCString dialogPassword;
        int passwordResult = KPasswordDialog::getPassword( dialogPassword, i18n("Password") );
        if ( passwordResult == KPasswordDialog::Accepted )
        {
          password = QString::fromUtf8( dialogPassword );
        }
      }
      hash = command[4] + password;
      KMD5 context( hash );
      // send the user verification
      sendCommand("USR", "MD5 S " + context.hexDigest() + "\r\n" );
*/
    }
    else if ( command[2] == "OK" )
    {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
      kdDebug() << "Notification: USR confirmation." << endl;
#endif
      // This is a confirmation message.  Get the user's friendly name from the message.
      QString friendlyName = command[4];
      removePercents( friendlyName );
      currentAccount_->setFriendlyName( friendlyName );

      emit statusMessage( i18n("Received user confirmation"), 2 );
      // Synchronize with the server (which will cause it to send the account data and contact lists)
      sendCommand( "SYN", "0\r\n" );
    }
  }
  else
  {
    kdDebug() << "Notification::gotUsr - WARNING - Current account is null." << endl;
  }
}



// Received version information
void MsnNotificationConnection::gotVer(const QStringList& command)
{
#ifdef KMESSTEST
  ASSERT( command[2] == "MSNP7" );
#endif
  // Send some fake info about the current version
  if ( currentAccount_ != 0 )
  {
    sendCommand( "CVR", "0x0409 winnt 5.1 i386 MSNMSGR 6.0.0250 MSMSGS " + currentAccount_->getHandle() + "\r\n" );
  }
}



// Received server transfer information
void MsnNotificationConnection::gotXfr(const QStringList& command)
{
  // Find out what the server type is.
  QString serverType = command[2];
  // Get the server ip and port (as server:port) from the message
  QString serverAndPort = command[3];
#ifdef KMESSTEST
  ASSERT( serverAndPort.contains(":") );
#endif
  // Get the server from that
  QString server = serverAndPort.left( serverAndPort.find(":") );
  QString portString = serverAndPort.right( serverAndPort.length() - serverAndPort.findRev(":") - 1 );
  // Convert the port to an integer
  bool goodPort;
  int port = portString.toInt( &goodPort );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Got " << serverType << " transfer to " << server << ":" << port << "." << endl;
#endif
  if ( !goodPort )
  {
    kdDebug() << "Notfication: Got bad port number : " << portString << "." << endl;
    return;
  }
  if ( serverType == "NS" )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification: Disconnecting from old server." << endl;
#endif
    // Disconnect from the existing server.
    disconnectFromServer( true );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification: Connecting to new server." << endl;
#endif
    // This is a notification server transfer.  Switch this socket to the given server.
    emit statusMessage( i18n("Transfer to notification server"), 2 );
    // Connect to the new server.
    if ( connectToServer( server, port ) )
    {
      // Send version information again
      //putVer();
    }
    else
    {
      kdDebug() << "Notification: WARNING - Couldn't connect to new server." << endl;
    }
  }
  else if ( serverType == "SB" )
  {
#ifdef KMESSTEST
    ASSERT( openChats_.count() > 0 );
    uint noChatsBefore = openChats_.count();
#endif
    // This is a switchboard server.  This message should be in response to
    //  a user-requested chat.  Look for the pending chat information.
    QString authorization = command[5];
    ChatInformation *chatInfo = openChats_.first();
    if ( chatInfo == 0 )
    {
      kdDebug() << "MsnNotificationConnection - Got switchboard, but there are no open chats." << endl;
      return;
    }
    // Update the chat information
    chatInfo->setServerInformation( server, port, authorization );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification signal a chat to " << server << ":" << port << endl;
#endif
    // Signal the chat
    emit startChat( chatInfo );
    // Remove the first chat from the open chats
    openChats_.remove( chatInfo );
#ifdef KMESSTEST
    ASSERT( openChats_.count() == ( noChatsBefore - 1 ) );
#endif
  }
}



// Initialize the object
bool MsnNotificationConnection::initialize(ContactListView *contactListView)
{
  if ( initialized_ )
  {
    kdDebug() << "Notification Connection already initialized!" << endl;
    return false;
  }
  if ( !MsnConnection::initialize() )
  {
    kdDebug() << "Notification Connection: Couldn't initialize parent." << endl;
    return false;
  }
  contactList_ = new ContactList();
  if ( !contactList_->initialize( contactListView ) )
  {
    kdDebug() << "MsnNotificationConnection couldn't initialize contact list." << endl;
    return false;
  }
  // Connect up the contact list
  connect( contactList_, SIGNAL(     requestAddContact(QString, QString, QString) ),
           this,         SLOT  (            addContact(QString, QString, QString) ) );
  connect( contactList_, SIGNAL(       requestAddGroup(QString)                     ),
           this,         SLOT  (              addGroup(QString)                     ) );
  connect( contactList_, SIGNAL(  requestRemoveContact(QString, QString, QString) ),
           this,         SLOT  (         removeContact(QString, QString, QString) ) );
  connect( contactList_, SIGNAL(    requestRemoveGroup(QString)                     ),
           this,         SLOT  (           removeGroup(QString)                     ) );
  connect( contactList_, SIGNAL(  requestRenameContact(QString, QString)           ),
           this,         SLOT  (    changeFriendlyName(QString, QString)           ) );
  connect( contactList_, SIGNAL(    requestRenameGroup(QString, QString)           ),
           this,         SLOT  (           renameGroup(QString, QString)           ) );
  connect( contactList_, SIGNAL(       chatWithContact(QString)                     ),
           this,         SLOT  (           requestChat(QString)                     ) );
  connect( contactList_, SIGNAL(        contactOffline(const Contact*)               ),
           this,         SLOT  ( forwardContactOffline(const Contact*)               ) );
  connect( contactList_, SIGNAL(         contactOnline(const Contact*)               ),
           this,         SLOT  (  forwardContactOnline(const Contact*)               ) );
  initialized_ = true;
  return initialized_;
}



// Open a connection to the server
bool MsnNotificationConnection::openConnection()
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Opening connection." << endl;
#endif
  if ( !initialized_ )
  {
    kdDebug() << "Notification: ERROR - Notification should be initialized before attempting to connect." << endl;
    return false;
  }

  bool connected;

  emit statusMessage( i18n( "Connecting..." ), 2 );

  // Attempt to connect to the main MSN server
  connected = connectToServer( "messenger.hotmail.com", 1863 );
  if ( connected )
  {
    // Start the login timer
    loginTimer_.start( timeoutDuration_, true );
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification: Socket started connecting OK.  Started. the timer." << endl;
#endif
  }
  else
  {
    kdDebug() << "Notification: Couldn't connect to the server." << endl;
  }
  return connected;
}



// Parse a regular command
void MsnNotificationConnection::parseCommand(const QStringList& command)
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Parsing command " << command[0] << "." << endl;
#endif
  if ( command[0] == "ADD" )
  {
    gotAdd( command );
  }
  else if ( command[0] == "ADG" )
  {
    gotAdg( command );
  }
  else if ( command[0] == "BLP" )
  {
    // Ignore this
  }
  else if ( command[0] == "BPR" )
  {
    // Ignore this
  }
  else if ( command[0] == "CHG" )
  {
    gotChg( command );
  }
  else if ( command[0] == "CHL" )
  {
    gotChl( command );
  }
  else if ( command[0] == "CVR" )
  {
    gotCvr( command );
  }
  else if ( command[0] == "FLN" )
  {
    gotFln( command );
  }
  else if ( command[0] == "GTC" )
  {
    // Ignore this
  }
  else if ( command[0] == "ILN" )
  {
    gotIln( command );
  }
  else if ( command[0] == "INF" )
  {
    gotInf( command );
  }
  else if ( command[0] == "LSG" )
  {
    gotLsg( command );
  }
  else if ( command[0] == "LST" )
  {
    gotLst( command );
  }
  else if ( command[0] == "NLN" )
  {
    gotNln( command );
  }
  else if ( command[0] == "OUT" )
  {
    gotOut( command );
  }
  else if ( command[0] == "PRP" )
  {
    // Ignore this
  }
  else if ( command[0] == "QRY" )
  {
    // Ignore this
  }
  else if ( command[0] == "REA" )
  {
    gotRea( command );
  }
  else if ( command[0] == "REG" )
  {
    gotReg( command );
  }
  else if ( command[0] == "REM" )
  {
    gotRem( command );
  }
  else if ( command[0] == "RMG" )
  {
    gotRmg( command );
  }
  else if ( command[0] == "RNG" )
  {
    gotRng( command );
  }
  else if ( command[0] == "SYN" )
  {
    gotSyn( command );
  }
  else if ( command[0] == "URL" )
  {
    gotUrl( command );
  }
  else if ( command[0] == "USR" )
  {
    gotUsr( command );
  }
  else if ( command[0] == "VER" )
  {
    gotVer( command );
  }
  else if ( command[0] == "XFR" )
  {
    gotXfr( command );
  }
  // Check error codes
  else if ( command[0] == "208" )
  {
    dotNetMessage( i18n("The username given was invalid") );
    closeConnection();
  }
  else if ( command[0] == "209" )
  {
    dotNetMessage( i18n("That name is invalid") );
//    closeConnection();
  }
  else if ( command[0] == "210" )
  {
    dotNetMessage( i18n("Your contact list is full") );
  }
  else if ( command[0] == "500" )
  {
    dotNetMessage( i18n("There was an internal server error") );
    closeConnection();
  }
  else if ( command[0] == "600" )
  {
    dotNetMessage( i18n("The server is busy") );
    closeConnection();
  }
  else if ( command[0] == "601" )
  {
    dotNetMessage( i18n("The server is unavailable") );
  }
  else if ( command[0] == "602" )
  {
    dotNetMessage( i18n("Peer notification server down") );
    closeConnection();
  }
  else if ( command[0] == "640" )
  {
    dotNetMessage( i18n("The server is going down soon") );
  }
  else if ( command[0] == "713" )
  {
    dotNetMessage( i18n("User too active") );
  }
  else if ( command[0] == "911" )
  {
    dotNetMessage( i18n("Authentication failed") );
    closeConnection();
  }
  else if ( command[0] == "913" )
  {
    KMessageBox::error( 0, i18n("You can't start a chat with someone while you are invisible.") );
  }
  else if ( command[0] == "919" )
  {
    dotNetMessage( i18n("Got mystery '919' error") );
    closeConnection();
  }
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  else
  {
    kdDebug() << "Notification: Command " <<  command[0] << " not supported." << endl;
  }
#endif
}



// Parse a message command
void MsnNotificationConnection::parseMessage(const QStringList& /*command*/, const MimeMessage &mainMessage)
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "MsnNotification - Got " << mainMessage.getSubValue("Content-Type") << "." << endl;
#endif
  QString contentType = mainMessage.getSubValue("Content-Type");
  MimeMessage message( mainMessage.getBody() );
#ifdef KMESSDEBUG_NOTIFICATION_MESSAGES
  mainMessage.print();
  kdDebug() << "MsnNotification - Message body as mime:" << endl;
  message.print();
#endif
  // Check the message type
  if ( contentType == "text/x-msmsgsprofile" )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "MsnNotification - Load profile information." << endl;
#endif
    currentAccount_->setAccountInformation( mainMessage.getValue("MSPAuth"),
                                            mainMessage.getValue("kv"),
                                            mainMessage.getValue("preferredEmail"),
                                            mainMessage.getValue("sid") );
  }
  else if ( contentType == "text/x-msmsgsinitialemailnotification" )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "MsnNotification - Load initial email information." << endl;
#endif
    currentAccount_->setInitialEmailInformation( message.getValue("Post-URL"),
                                                 message.getValue("Inbox-Unread").toInt(),
                                                 message.getValue("Folders-Unread").toInt() );
  }
  else if ( contentType == "text/x-msmsgsemailnotification" )
  {
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
    kdDebug() << "Notification: An email was received from \"" << message.getValue("From") << "\" with the subject \"" << message.getValue("Subject") << "\" in the folder \"" << message.getValue("Dest-Folder") << "\"." << endl;
#endif
    QString destinationDir;
    // If a message delta is to the active folder, emails were moved into it, so use a negative deletion
    if ( message.getValue("Dest-Folder") == "ACTIVE" )
    {
      currentAccount_->changeNoEmails( 1 );
      destinationDir = i18n("in your inbox");
    }
    else
    {
      destinationDir = i18n("in another folder");
    }

    emit newEmail( message.decodeRFC2047String( message.getValue("From").data() ),
                   message.decodeRFC2047String( message.getValue("Subject").data() ),
                   destinationDir,
                   message.getValue("id"),
                   message.getValue("Message-URL"),
                   message.getValue("Post-URL") );
  }
  else if ( contentType == "text/x-msmsgsactivemailnotification" )
  {
    int messageDelta = message.getValue("Message-Delta").toInt();

    // If a message delta is from the active folder, emails were deleted
    if ( message.getValue("Src-Folder") == "ACTIVE" )
    {
      currentAccount_->changeNoEmails( -1 * messageDelta );
    }
    // If a message delta is to the active folder, emails were moved into it
    else if ( message.getValue("Dest-Folder") == "ACTIVE" )
    {
      currentAccount_->changeNoEmails( messageDelta );
    }
  }
}



// Show a message that the proxy failed
void MsnNotificationConnection::proxyFailed()
{
  QString message;
  // Show a message to the user
  message = i18n("There was a problem connecting to the proxy.");
  KMessageBox::error( 0, message );
  // Close the connection
  closeConnection();
}



// Send the version command
void MsnNotificationConnection::putVer()
{
  // Send the version command for MSN 9
  sendCommand( "VER", "MSNP9 MSNP8 CVR0\r\n" );
}



// Remove a contact from the given list or, if applicable, group
void MsnNotificationConnection::removeContact(QString handle, QString list, QString groupId)
{
#ifdef KMESSTEST
  bool goodGroupId;
  groupId.toInt( &goodGroupId );
  ASSERT( ( goodGroupId ) || ( list == "FL" ) );
  ASSERT( ( list == "FL" ) || ( list == "AL" ) || ( list == "BL" ) );
#endif
  QString command;
  command = list + " " + handle;
  if ( list == "FL" )
  {
    command += " " + groupId;
  }
  command += "\r\n";
  sendCommand( "REM", command );
}



// Remove a group
void MsnNotificationConnection::removeGroup(QString id)
{
#ifdef KMESSTEST
  bool goodGroupId;
  id.toInt( &goodGroupId );
  ASSERT( goodGroupId );
#endif
  sendCommand( "RMG", id + "\r\n" );
}



// Rename a group
void MsnNotificationConnection::renameGroup(QString id, QString newName)
{
#ifdef KMESSTEST
  bool goodGroupId;
  id.toInt( &goodGroupId );
  ASSERT( goodGroupId );
#endif
  addPercents( newName );
  sendCommand( "REG", id + " " + newName + " 0\r\n");
}



// Request a chat with a contact
void MsnNotificationConnection::requestChat(QString handle)
{
  // Create a new chat information object and store it so that it can be
  //  given the switchboard server info when it's returned
  openChats_.append( new ChatInformation( handle, contactList_ ) );
  // Save the contact information?
  sendCommand( "XFR", "SB\r\n" );
}



// Save contact list properties
void MsnNotificationConnection::saveProperties(KConfig *config)
{
#ifdef KMESSDEBUG_NOTIFICATION
  kdDebug() << "Notification - saving properties" << endl;
#endif
  contactList_->saveProperties( config );
}



// Have the contact list show the new group dialog
void MsnNotificationConnection::showNewGroupDialog()
{
#ifdef KMESSTEST
  ASSERT( contactList_ != 0 );
#endif
  if ( contactList_ != 0 )
  {
    contactList_->newGroupDialog();
  }
}



// The SSL login failed
void MsnNotificationConnection::sslLoginFailed()
{
#ifdef KMESSDEBUG_NOTIFICATION_GENERAL
  kdDebug() << "Notification: Login SSL login failed." << endl;
#endif
  QString message;
  // Delete the login handler
  delete sslLoginHandler_;
  sslLoginHandler_ = 0;
  // Show a message to the user
  message = i18n("The SSL portion of the login failed.");
  KMessageBox::error( 0, message );
  closeConnection();
}



// The SSL login succeeded
void MsnNotificationConnection::sslLoginSucceeded(QString authentication)
{
#ifdef KMESSDEBUG_NOTIFICATION
  kdDebug() << "Notification - Authentication is " << authentication << endl;
#endif
  // Delete the login handler
  delete sslLoginHandler_;
  sslLoginHandler_ = 0;
  sendCommand( "USR", "TWN S " + authentication + "\r\n" );
}

