/***************************************************************************
                          msnswitchboardconnection.h  -  description
                             -------------------
    begin                : Fri Jan 24 2003
    copyright            : (C) 2003 by Mike K. Bennett
    email                : mkb137b@hotmail.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MSNSWITCHBOARDCONNECTION_H
#define MSNSWITCHBOARDCONNECTION_H

#include <qfont.h>
#include <qstringlist.h>
#include <qvaluelist.h>

#include "msnconnection.h"

// Forward declarations
class Application;
class ChatInformation;
class KMessTest;
class MimeMessage;

/**A connection to a switchboard server, over which chats are conducted.
  *@author Mike K. Bennett
  */

class MsnSwitchboardConnection : public MsnConnection
{
  Q_OBJECT

  friend class KMessTest;

  public:
    // The supported applications
    enum ApplicationType { GNOMEMEETING, SENDFILE, RECEIVEFILE, MSNREMOTEDESKTOP, KDEREMOTEDESKTOP };
    // The constructor
  	                     MsnSwitchboardConnection();
    // The destructor
  	                    ~MsnSwitchboardConnection();
    // Close the connection
    void                 closeConnection();
    // Get the switchboard to make a list of the contacts in the chat
    QString              getCaption();
    // Make a list of the contacts in the chat
    QStringList          getContactList() const;
    // Initialize the object
    bool                 initialize();
    // Invite a contact into the chat
    void                 inviteContact(QString handle);
    // Check if only the given contact is in the chat
    bool                 isExclusiveChatWithContact(const QString& handle) const;
    // Start an application
    void                 startApp(ApplicationType type);
    // Start a chat
    void                 startChat(ChatInformation *chatInfo);

  public slots: // Public slots
     // Give the application with the given cookie the command.
    void                 giveAppCommand(QString cookie, QString command);
    // Send a message to the contact(s)
    void                 sendMessage(QString text);
    // The user is typing so send a typing message
    void                 userIsTyping();

  protected slots: // Protected slots
    // The socket connected, so send the version command.
    void                 connectionSuccess();

  private: // Private methods
    // Add a contact to the list of contacts in the chat
    void                 addContactToList(const QString& handle);
    // Make the neccessary application connections.
    void                 connectApplication(Application *app);
    // Do what's required when a contact comes online
    void                 contactOnline(QString& handle, QString& friendlyName);
    // Convert an html format (#RRGGBB) color to an msn format (BBGGRR) color
    void                 convertHtmlColorToMsnColor(QString &color) const;
    // Convert and msn format color (BBGGRR) to an html format (#RRGGBB) color
    void                 convertMsnColorToHtmlColor(QString &color) const;
    // Create an application of the given type
    Application*         createApplication(ApplicationType type);
    // Return the application with the given cookie.
    Application*         getApplication(QString cookie) const;
    // Get a contact's friendly name from the contact list
    QString              getContactFriendlyNameByHandle(const QString &handle);
    // Get the font and color from a message
    void                 getFontAndColorFromMessage(QFont &font, QString &color, const MimeMessage &message) const;
    // Get a font from the message
    void                 getFontFromMessageFormat(QFont &font, const MimeMessage& message) const;
    // Get an html font color from the message
    void                 getFontColorFromMessageFormat(QString &color, const MimeMessage& message) const;
    // Extract some information from a font message (i.e. "EF=UIB;" extract the "UIB" )
    QString              getFontValueFromMessageFormat(const QString& field, const QString& message) const;
    // Return a properly created message head
    QString              getMessageHead();
    // Return the "user is typing" message
    QString              getTypingMessage() const;
    // Received a positive delivery message.
    void                 gotAck(const QStringList& command);
    // Received notification that a contact is no longer in session.
    void                 gotBye(const QStringList& command);
    // Received the initial roster information for new contacts joining a session.
    void                 gotIro(const QStringList& command);
    // Received notification of a new client in the session.
    void                 gotJoi(const QStringList& command);
    // Received a negative acknowledgement of the receipt of a message.
    void                 gotNak(const QStringList& command);
    // Received notification of the termination of a client-server session.
    void                 gotOut(const QStringList& command);
    // Received a client-server authentication message.
    void                 gotUsr(const QStringList& command);
    // Check if a certain contact is in the chat
    bool                 isContactInChat(const QString& handle) const;
    // Parse an application message
    void                 parseAppMessage(const MimeMessage &message);
    // Parse a regular command
    void                 parseCommand(const QStringList& command);
    // Parse a message command
    void                 parseMessage(const QStringList& command, const MimeMessage &message);
    // Send a reject message for the given application
    void                 rejectApplication(const QString& cookie);
    // Remove a contact from the list of contacts in the chat
    void                 removeContactFromList(const QString& handle);
    // Send messages that weren't sent because a contact had to be re-called
    void                 sendPendingMessages();
    // Show a "Cancelled" message to the user
    void                 showAppCancelled();
    // Store a message for later acknowledgement
    void                 storeMessageForAcknowledgement(const int& ack, const QString& text);

  private slots: // Private slots
    // Forward an application message by reemitting it as a signal.
    void                 forwardAppMessage(QString message);
    // Send an application message to the server.
    void                 putAppMsg(QString message);
    // Remove and delete the given application.
    void                 terminateApplication(QString cookie);

  private: // Private attributes
    // String lists to store messages and their ack numbers for confirmation of receipt
    QValueList<QString>  acks_, messages_;
    // A list of pointers to the open applications.
    QPtrList<Application> applications_;
    // The stored chat information
    ChatInformation     *chatInfo_;
    // Whether there are contacts in this chat, despite the default one that is left in contactsInChat_
    bool                 contactsConnected_;
    // A list of the contacts in the chat
    QValueList<QString>  contactsInChat_;
    // Whether or not the object has been initialized
    bool                 initialized_;
    // A list of messages that should be sent to the contact when connection is reestablished
    QValueList<QString>  pendingMessages_;

  signals: // Public signals
    // Signal that a contact joined (or was already in) the chat
    void                 contactJoinedChat( QString handle, QString friendlyName );
    // Signal that a contact left the chat
    void                 contactLeftChat( QString handle );
    // Signal that a contact is typing
    void                 contactTyping( QString handle, QString friendlyName );
    // Signal a chat message
    void                 chatMessage( QString handle, QString name, QString text, QFont font, QString color, QString nameColor );
    // Send a status message about the connection
    void                 systemMessage( QString message );
    // Show a forwarded application message.
    void                 showAppMessage(QString message);
};

#endif
